from __future__ import annotations

import asyncio
from typing import (
    Any,
    AsyncIterable,
    AsyncIterator,
    Awaitable,
    Callable,
    Deque,
    Generic,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

T = TypeVar("T")
R = TypeVar("R")


async def consume(iterable: AsyncIterable[Any]) -> None:
    async for _ in iterable:
        pass


class BoostExecutor:
    """BoostExecutor implements a concurrent.futures-like interface for running async tasks.

    The idea is: you want to run a limited number of tasks at a time. Moreover, if you have spare
    capacity, you want to redistribute it to tasks you're running. For instance, when downloading
    chunks of a file, we could use spare capacity to proactively download future chunks. That is,
    instead of the following serial code:
    ```
    for chunk in iterable_of_chunks:
        data = await download_func(chunk)
        write(data)
    ```
    You can instead do:
    ```
    async with BoostExecutor(4) as executor:
        async for data in executor.map_ordered(download_func, iterable_of_chunks):
            write(data)
    ```

    The number supplied to BoostExecutor is the number of units of concurrency you wish to maintain.
    That is, using ``BoostExecutor(1)`` would make the above example function approximately similar
    to the serial code.

    """

    def __init__(self, concurrency: int) -> None:
        assert concurrency > 0
        # Okay, so this is a little tricky. We take away one unit of concurrency now, and give it
        # back when you iterate over a boostable. You can think of concurrency - 1 as the number of
        # units of "background" concurrency. When our "foreground" starts iterating over a
        # boostable, for instance, with an async for loop, it donates one unit of concurrency to the
        # cause for the duration of the iteration.
        # Why this rigmarole? To avoid deadlocks on reentrancy. If a boostable spawns more work on
        # the same executor and then blocks on that work, it's holding a unit of concurrency while
        # trying to acquire another one. If that happens enough times, we deadlock. For examples,
        # see test_composition_nested_(un)?ordered.
        # To solve this, we treat Boostable.__aiter__ as an indicator that we are entrant. We then
        # donate a unit of concurrency by releasing the semaphore. Conceptually, this is our
        # "foreground" or "re-entrant" unit of concurrency. We give this back when the Boostable is
        # done iterating. We could be more granular and do this on every Boostable.__anext__, but
        # that's a lot more overhead, makes iteration less predictable and can result in what feels
        # like unnecessary blocking. We could also do something complicated using contextvars, but
        # that would have to be something complicated and would have to use contextvars.
        self.semaphore = asyncio.Semaphore(concurrency - 1)
        self.boostables: Deque[Boostable[Any, Any]] = Deque()

        self.waiter: Optional[asyncio.Future[None]] = None
        self.runner: Optional[asyncio.Task[None]] = None
        self.shutdown: bool = False

    async def __aenter__(self) -> BoostExecutor:
        self.runner = asyncio.create_task(self.run())
        return self

    async def __aexit__(
        self, exc_type: Optional[Type[BaseException]], exc_value: Any, traceback: Any
    ) -> None:
        self.shutdown = True
        if exc_type:
            # If there was an exception, let it propagate and don't block on the runner exiting.
            # Also cancel the runner.
            assert self.runner is not None
            self.runner.cancel()
            return
        self.notify_runner()
        assert self.runner is not None
        await self.runner

    def map_ordered(
        self, func: Callable[[T], Awaitable[R]], iterable: BoostUnderlying[T]
    ) -> OrderedBoostable[T, R]:
        ret = OrderedBoostable(func, iterable, self.semaphore)
        self.boostables.appendleft(ret)
        self.notify_runner()
        return ret

    def map_unordered(
        self, func: Callable[[T], Awaitable[R]], iterable: BoostUnderlying[T]
    ) -> UnorderedBoostable[T, R]:
        ret = UnorderedBoostable(func, iterable, self.semaphore)
        self.boostables.appendleft(ret)
        self.notify_runner()
        return ret

    def notify_runner(self) -> None:
        if self.waiter and not self.waiter.done():
            self.waiter.set_result(None)

    async def run(self) -> None:
        loop = asyncio.get_event_loop()
        exhausted_boostables: List[Boostable[Any, Any]] = []
        not_ready_boostables: Deque[Boostable[Any, Any]] = Deque()

        MIN_TIMEOUT = 0.01
        MAX_TIMEOUT = 0.1
        timeout = MIN_TIMEOUT

        while True:
            await self.semaphore.acquire()
            self.semaphore.release()

            while self.boostables:
                # We round robin the boostables until they're either all exhausted or not ready
                task = self.boostables[0].provide_boost()
                if isinstance(task, NotReady):
                    not_ready_boostables.append(self.boostables.popleft())
                    continue
                if isinstance(task, Exhausted):
                    exhausted_boostables.append(self.boostables.popleft())
                    continue
                assert isinstance(task, asyncio.Task)

                await asyncio.sleep(0)
                self.boostables.rotate(-1)
                if self.semaphore.locked():
                    break
            else:
                self.boostables = not_ready_boostables
                not_ready_boostables = Deque()

            if self.semaphore.locked():
                # If we broke out of the inner loop due to a lack of available concurrency, go to
                # the top of the outer loop and wait for concurrency
                continue

            if self.shutdown and not self.boostables:
                # If we've been told to shutdown and we have nothing more to boost, exit
                break

            self.waiter = loop.create_future()
            try:
                # If we still have boostables (because they're currently not ready), timeout in
                # case they become ready
                # Otherwise, (since we have available concurrency), wait for us to be notified in
                # case of more work
                await asyncio.wait_for(self.waiter, timeout if self.boostables else None)
                # If we were waiting for boostables to become ready, increase how long we'd wait
                # the next time. Otherwise, reset the timeout duration.
                timeout = min(MAX_TIMEOUT, timeout * 2) if self.boostables else MIN_TIMEOUT
            except asyncio.TimeoutError:
                pass
            self.waiter = None

        # Boostables are exhausted if their underlying iterable has been consumed, but there still
        # might be tasks running. Wait for them to finish.
        for boostable in exhausted_boostables:
            await boostable.wait()

        # Yield, so that iterations over boostables are likely to have finished by the time we exit
        # the BoostExecutor context
        await asyncio.sleep(0)


class Exhausted:
    pass


class NotReady:
    pass


class Boostable(Generic[T, R]):
    """A Boostable represents some operation that could make use of additional concurrency.

    You probably shouldn't use these classes directly, instead, instantiate them via the ``map_*``
    methods of the executor that will provide these Boostables with additional concurrency.

    We represent boostable operations as an async function that is called on each element of some
    underlying iterable. To boost this operation, we dequeue another element from the underlying
    iterable and create an asyncio task to perform the function call over that element.

    Boostables also compose; that is, the underlying iterable can be another Boostable. In this
    case, we quickly (aka non-blockingly) check whether the underlying Boostable has an element
    ready. If so, we dequeue it, just as we would with an iterable, if not, we forward the boost
    we received to the underlying Boostable.

    You can treat Boostables as async iterators. An OrderedBoostable will yield results
    corresponding to the order of the underlying iterable, whereas an UnorderedBoostable will
    yield results based on what finishes first.

    Note that both kinds of Boostables will, of course, dequeue from the underlying iterable in
    whatever order the underlying provides. This means we'll start tasks in order (but not
    necessarily finish them in order, which is kind of the point). In some places, we use this
    fact to increment a shared index; it's important we do ordered shared memory things like that
    before any awaits.

    """

    def __init__(
        self,
        func: Callable[[T], Awaitable[R]],
        iterable: BoostUnderlying[T],
        semaphore: asyncio.Semaphore,
    ) -> None:
        if not isinstance(iterable, (Iterator, Boostable, EagerAsyncIterator)):
            raise ValueError(
                "Underlying iterable must be an Iterator, Boostable or EagerAsyncIterator"
            )

        async def wrapper(arg: T) -> R:
            async with semaphore:
                return await func(arg)

        self.func = wrapper
        self.iterable = iterable
        self.semaphore = semaphore

    def provide_boost(self) -> Union[NotReady, Exhausted, asyncio.Task[Any]]:
        """Start an asyncio task to help speed up this boostable.

        Raises NotReady if we can't make use of a boost currently.
        Raises Exhausted if we're done. This causes the BoostExecutor to stop attempting to
        provide boosts to this Boostable.

        """
        arg: T
        result: Union[NotReady, Exhausted, T]
        if isinstance(self.iterable, Boostable):
            result = self.iterable.dequeue()
            if isinstance(result, NotReady):
                # the underlying boostable doesn't have anything ready for us
                # so forward the boost to the underlying boostable
                return self.iterable.provide_boost()
            arg = result
        elif isinstance(self.iterable, EagerAsyncIterator):
            result = self.iterable.dequeue()
            if isinstance(result, (NotReady, Exhausted)):
                return result
            arg = result
        elif isinstance(self.iterable, Iterator):
            try:
                arg = next(self.iterable)
            except StopIteration:
                return Exhausted()
        else:
            raise AssertionError

        return self.enqueue(arg)

    async def wait(self) -> None:
        """Wait for all tasks in this boostable to finish."""
        raise NotImplementedError

    def enqueue(self, arg: T) -> asyncio.Task[R]:
        """Start and return an asyncio task based on an element from the underlying."""
        raise NotImplementedError

    def dequeue(self) -> Union[NotReady, R]:
        """Non-blockingly dequeue a result we have ready.

        Returns NotReady if it's not ready to dequeue.

        """
        raise NotImplementedError

    async def blocking_dequeue(self) -> R:
        """Dequeue a result, waiting if necessary."""
        raise NotImplementedError

    def __aiter__(self) -> AsyncIterator[R]:
        async def iterator() -> AsyncIterator[R]:
            try:
                self.semaphore.release()
                while True:
                    yield await self.blocking_dequeue()
            except StopAsyncIteration:
                pass
            finally:
                await self.semaphore.acquire()

        return iterator()


class OrderedBoostable(Boostable[T, R]):
    def __init__(
        self,
        func: Callable[[T], Awaitable[R]],
        iterable: BoostUnderlying[T],
        semaphore: asyncio.Semaphore,
    ) -> None:
        super().__init__(func, iterable, semaphore)
        self.buffer: Deque[asyncio.Task[R]] = Deque()

    async def wait(self) -> None:
        if self.buffer:
            await asyncio.wait(self.buffer)

    def enqueue(self, arg: T) -> asyncio.Task[R]:
        task = asyncio.create_task(self.func(arg))
        self.buffer.append(task)
        return task

    def dequeue(self) -> Union[NotReady, R]:
        if not self.buffer or not self.buffer[0].done():
            return NotReady()
        task = self.buffer.popleft()
        return task.result()

    async def blocking_dequeue(self) -> R:
        while True:
            if not self.buffer:
                arg = await next_underlying(self.iterable)
                self.enqueue(arg)
            ret = self.dequeue()
            if not isinstance(ret, NotReady):
                return ret
            # note that dequeues are racy, so we can't return the result of self.buffer[0]
            await self.buffer[0]


class UnorderedBoostable(Boostable[T, R]):
    def __init__(
        self,
        func: Callable[[T], Awaitable[R]],
        iterable: BoostUnderlying[T],
        semaphore: asyncio.Semaphore,
    ) -> None:
        super().__init__(func, iterable, semaphore)
        self.buffer: Set[asyncio.Task[R]] = set()
        self.waiter: Optional[asyncio.Future[asyncio.Task[R]]] = None

    def done_callback(self, task: asyncio.Task[R]) -> None:
        if self.waiter and not self.waiter.done():
            self.waiter.set_result(task)

    async def wait(self) -> None:
        if self.buffer:
            await asyncio.wait(self.buffer)

    def enqueue(self, arg: T) -> asyncio.Task[R]:
        task = asyncio.create_task(self.func(arg))
        self.buffer.add(task)
        task.add_done_callback(self.done_callback)
        return task

    def dequeue(self, hint: Optional[asyncio.Task[R]] = None) -> Union[NotReady, R]:
        # hint is a task that we suspect is dequeuable, which allows us to skip the linear check
        # against all outstanding tasks in the happy case.
        if hint is not None and hint in self.buffer and hint.done():
            task = hint
        else:
            try:
                task = next(t for t in self.buffer if t.done())
            except StopIteration:
                return NotReady()
        self.buffer.remove(task)
        return task.result()

    async def blocking_dequeue(self) -> R:
        task = None
        loop = asyncio.get_event_loop()
        while True:
            if not self.buffer:
                arg = await next_underlying(self.iterable)
                task = self.enqueue(arg)
            ret = self.dequeue(hint=task)
            if not isinstance(ret, NotReady):
                return ret
            # note that dequeues are racy, so the task we get from self.waiter may already have been
            # dequeued. in the happy case, however, it's probably the task we want to return, so we
            # use it as the hint for the next dequeue.
            self.waiter = loop.create_future()
            task = await self.waiter
            self.waiter = None


class EagerAsyncIterator(Generic[T]):
    """Like an AsyncIterator, but eager.

    AsyncIterators will only start computing their next element when awaited on.
    This allows an AsyncIterator to serve as the underlying iterable for a Boostable.
    Note a) this takes away the natural backpressure, b) instantiating one of these will use
    one unit of concurrency until it's exhausted which is not accounted for by a BoostExecutor.

    """

    def __init__(self, iterable: AsyncIterator[T]):
        async def eagerify(it: AsyncIterator[T]) -> Tuple[T, asyncio.Task[Any]]:
            val = await it.__anext__()
            next_task = asyncio.create_task(eagerify(it))
            return val, next_task

        self.next_task = asyncio.create_task(eagerify(iterable))

    def dequeue(self) -> Union[NotReady, Exhausted, T]:
        if not self.next_task.done():
            return NotReady()
        try:
            ret, self.next_task = self.next_task.result()
            return ret
        except StopAsyncIteration:
            return Exhausted()

    def __aiter__(self) -> AsyncIterator[T]:
        return self

    async def __anext__(self) -> T:
        while not self.next_task.done():
            # note that dequeues are racy, so we can't just return the result of self.next_task
            await self.next_task
        ret, self.next_task = self.next_task.result()
        return ret


# see docstring of Boostable
BoostUnderlying = Union[Iterator[T], Boostable[Any, T], EagerAsyncIterator[T]]


async def next_underlying(iterable: BoostUnderlying[T]) -> T:
    """Like ``next``, but abstracts over a BoostUnderlying."""
    if isinstance(iterable, Boostable):
        return await iterable.blocking_dequeue()
    if isinstance(iterable, EagerAsyncIterator):
        return await iterable.__anext__()
    if isinstance(iterable, Iterator):
        try:
            return next(iterable)
        except StopIteration:
            raise StopAsyncIteration
    raise AssertionError


async def iter_underlying(iterable: BoostUnderlying[T]) -> AsyncIterator[T]:
    """Like ``iter``, but abstracts over a BoostUnderlying."""
    if isinstance(iterable, (Boostable, EagerAsyncIterator)):
        async for x in iterable:
            yield x
    elif isinstance(iterable, Iterator):
        for x in iterable:
            yield x
    else:
        raise AssertionError



from __future__ import annotations

import asyncio
import base64
import hashlib
import os
import random
from typing import List, Mapping, Optional, Tuple, Union

from .boost import BoostExecutor, BoostUnderlying, consume, iter_underlying
from .delete import remove
from .path import AzurePath, BasePath, CloudPath, GooglePath, LocalPath, exists, pathdispatch
from .read import ByteRange
from .request import Request, azurify_request, googlify_request

AZURE_BLOCK_COUNT_LIMIT = 50000

# ==============================
# write_single
# ==============================


@pathdispatch
async def write_single(path: Union[BasePath, str], data: bytes, overwrite: bool = False) -> None:
    """Write the given stream to ``path``.

    :param path: The path to write to.
    :param data: The data to write.
    :param overwrite: If False, raises if the path already exists.

    """
    raise ValueError(f"Unsupported path: {path}")


@write_single.register  # type: ignore
async def _azure_write_single(path: AzurePath, data: bytes, overwrite: bool = False) -> None:
    if not overwrite:
        if await exists(path):
            raise FileExistsError(path)

    request = await azurify_request(
        Request(
            method="PUT",
            url=path.format_url("https://{account}.blob.core.windows.net/{container}/{blob}"),
            data=data,
            headers={"x-ms-blob-type": "BlockBlob"},
            success_codes=(201,),
        )
    )
    await request.execute_reponseless()


@write_single.register  # type: ignore
async def _google_write_single(path: GooglePath, data: bytes, overwrite: bool = False) -> None:
    if not overwrite:
        if await exists(path):
            raise FileExistsError(path)

    request = await googlify_request(
        Request(
            method="POST",
            url=path.format_url(
                "https://storage.googleapis.com/upload/storage/v1/b/{bucket}/o?uploadType=media&name={blob}"
            ),
            data=data,
            headers={"Content-Type": "application/octet-stream"},
        )
    )
    await request.execute_reponseless()


@write_single.register  # type: ignore
async def _local_write_single(path: LocalPath, data: bytes, overwrite: bool = False) -> None:
    if not overwrite:
        if await exists(path):
            raise FileExistsError(path)

    os.makedirs(path.parent, exist_ok=True)
    with open(path, mode="wb") as f:
        f.write(data)


# ==============================
# write_stream
# ==============================


@pathdispatch
async def write_stream(
    path: Union[BasePath, str],
    stream: BoostUnderlying[bytes],
    executor: BoostExecutor,
    overwrite: bool = False,
) -> None:
    """Write the given stream to ``path``.

    :param path: The path to write to.
    :param executor: An executor.
    :param stream: The stream of bytes to write. Note the chunking of stream also determines how
        we chunk the writes. Writes to Google Cloud must be chunked in multiples of 256 KB.
    :param overwrite: If False, raises if the path already exists.

    """
    raise ValueError(f"Unsupported path: {path}")


@write_stream.register  # type: ignore
async def _azure_write_stream(
    path: AzurePath,
    stream: BoostUnderlying[bytes],
    executor: BoostExecutor,
    overwrite: bool = False,
) -> None:
    if overwrite:
        # if the existing blob type is not compatible with the block blob we are about to write we
        # have to delete the file before writing our block blob or else we will get a 409 error when
        # putting the first block
        # if the existing blob is compatible, then in the event of multiple concurrent writers we
        # run the risk of ending up with uncommitted blocks, which could hit the uncommitted block
        # limit. rather than deal with that, just remove the file before writing which will clear
        # all uncommitted blocks
        # we could have a more elaborate upload system that does a write, then a copy, then a delete
        # but it's not obvious how to ensure that the temporary file is deleted without creating a
        # lifecycle rule on each container
        # TODO: blobfile has made some changes here, consider pulling them in
        try:
            await remove(path)
        except FileNotFoundError:
            pass
    else:
        if await exists(path):
            raise FileExistsError(path)

    upload_id = random.randint(0, 2 ** 47 - 1)
    block_index = 0
    md5 = hashlib.md5()

    async def upload_chunk(chunk: bytes) -> None:
        # mutating an index in the outer scope is a little sketchy, but it works out if we do it
        # before we await
        nonlocal block_index
        block_id = _get_block_id(upload_id, block_index)
        block_index += 1
        # https://docs.microsoft.com/en-us/rest/api/storageservices/put-block-list#remarks
        assert block_index < AZURE_BLOCK_COUNT_LIMIT
        md5.update(chunk)
        await _azure_put_block(path, block_id, chunk)

    await consume(executor.map_ordered(upload_chunk, stream))

    # azure does not calculate md5s for us, we have to do that manually
    # https://blogs.msdn.microsoft.com/windowsazurestorage/2011/02/17/windows-azure-blob-md5-overview/
    headers = {"x-ms-blob-content-md5": base64.b64encode(md5.digest()).decode("utf8")}
    blocklist = [_get_block_id(upload_id, i) for i in range(block_index)]
    await _azure_put_block_list(path, blocklist, headers=headers)


@write_stream.register  # type: ignore
async def _google_write_stream(
    path: GooglePath,
    stream: BoostUnderlying[bytes],
    executor: BoostExecutor,
    overwrite: bool = False,
) -> None:
    if not overwrite:
        if await exists(path):
            raise FileExistsError(path)

    upload_url = await _google_start_resumable_upload(path)

    offset = 0
    is_finalised = False
    total_size = "*"
    # Unfortunately, GCS doesn't allow us to upload these chunks in parallel.
    async for chunk in iter_underlying(stream):
        start = offset
        offset += len(chunk)
        end = offset

        # GCS requires resumable uploads to be chunked in multiples of 256 KB, except for the last
        # chunk. If you upload a chunk with another size you get an HTTP 400 error, unless you tell
        # GCS that it's the last chunk. Since our interface doesn't allow us to know whether or not
        # a given chunk is actually the last chunk, we go ahead and assume that it is if it's an
        # invalid chunk size. If we receive multiple chunks of invalid chunk size, we throw an
        # error.
        should_finalise = len(chunk) % (256 * 1024) != 0
        if should_finalise:
            if is_finalised:
                raise ValueError(
                    "The upload was already finalised. A likely cause is the given stream was "
                    "chunked incorrectly. Uploads to Google Cloud need to be chunked in "
                    "multiples of 256 KB (except for the last chunk)."
                )
            total_size = str(end)
            is_finalised = True

        request = await googlify_request(
            Request(
                method="PUT",
                url=upload_url,
                data=chunk,
                headers={
                    "Content-Type": "application/octet-stream",
                    "Content-Range": f"bytes {start}-{end-1}/{total_size}",
                },
                success_codes=(200, 201) if is_finalised else (308,),
            )
        )
        await request.execute_reponseless()

    if not is_finalised:
        await _google_finalise_upload(upload_url, total_size=offset)


@write_stream.register  # type: ignore
async def _local_write_stream(
    path: LocalPath,
    stream: BoostUnderlying[bytes],
    executor: BoostExecutor,
    overwrite: bool = False,
) -> None:
    if not overwrite:
        if await exists(path):
            raise FileExistsError(path)

    os.makedirs(path.parent, exist_ok=True)
    # TODO: evaluate whether running in executor actually helps
    loop = asyncio.get_event_loop()
    with open(path, mode="wb") as f:
        async for data in iter_underlying(stream):
            # f.write(data)
            await loop.run_in_executor(None, f.write, data)


# ==============================
# write_stream_unordered
# ==============================


@pathdispatch
async def write_stream_unordered(
    path: Union[CloudPath, str],
    stream: BoostUnderlying[Tuple[bytes, ByteRange]],
    executor: BoostExecutor,
    overwrite: bool = False,
) -> None:
    """Write the given stream to ``path``.

    :param path: The path to write to.
    :param executor: An executor.
    :param stream: The stream of bytes to write, along with what range of bytes each chunk
        corresponds to. Note the chunking of stream also determines how we chunk the writes.
    :param overwrite: If False, raises if the path already exists.

    """
    raise ValueError(f"Unsupported path: {path}")


@write_stream_unordered.register  # type: ignore
async def _azure_write_stream_unordered(
    path: AzurePath,
    stream: BoostUnderlying[Tuple[bytes, ByteRange]],
    executor: BoostExecutor,
    overwrite: bool = False,
) -> None:
    # TODO: this doesn't upload an md5...
    if overwrite:
        try:
            await remove(path)
        except FileNotFoundError:
            pass
    else:
        if await exists(path):
            raise FileExistsError(path)

    upload_id = random.randint(0, 2 ** 47 - 1)
    iter_index = 0
    block_list = []

    async def upload_chunk(chunk_byte_range: Tuple[bytes, ByteRange]) -> None:
        # mutating an index in the outer scope is a little sketchy, but it works out if we do it
        # before we await
        nonlocal iter_index
        block_id = _get_block_id(upload_id, iter_index)
        chunk, byte_range = chunk_byte_range
        block_list.append((byte_range[0], iter_index))

        iter_index += 1
        # https://docs.microsoft.com/en-us/rest/api/storageservices/put-block-list#remarks
        assert iter_index < AZURE_BLOCK_COUNT_LIMIT

        await _azure_put_block(path, block_id, chunk)

    await consume(executor.map_unordered(upload_chunk, stream))

    # sort by start byte so the blocklist is ordered correctly
    block_list.sort()
    await _azure_put_block_list(path, [_get_block_id(upload_id, index) for _, index in block_list])


@write_stream_unordered.register  # type: ignore
async def _google_write_stream_unordered(
    path: GooglePath,
    stream: BoostUnderlying[Tuple[bytes, ByteRange]],
    executor: BoostExecutor,
    overwrite: bool = False,
) -> None:
    raise NotImplementedError


# ==============================
# helpers
# ==============================


def _get_block_id(upload_id: int, index: int) -> str:
    assert index < 2 ** 17
    id_plus_index = (upload_id << 17) + index
    assert id_plus_index < 2 ** 64
    return base64.b64encode(id_plus_index.to_bytes(8, byteorder="big")).decode("utf8")


async def _azure_put_block(path: AzurePath, block_id: str, chunk: bytes) -> None:
    request = await azurify_request(
        Request(
            method="PUT",
            url=path.format_url("https://{account}.blob.core.windows.net/{container}/{blob}"),
            params=dict(comp="block", blockid=block_id),
            data=chunk,
            success_codes=(201,),
        )
    )
    await request.execute_reponseless()


async def _azure_put_block_list(
    path: AzurePath, block_list: List[str], headers: Optional[Mapping[str, str]] = None
) -> None:
    if headers is None:
        headers = {}
    request = await azurify_request(
        Request(
            method="PUT",
            url=path.format_url("https://{account}.blob.core.windows.net/{container}/{blob}"),
            headers=headers,
            params=dict(comp="blocklist"),
            data={"BlockList": {"Latest": block_list}},
            success_codes=(201,),
        )
    )
    await request.execute_reponseless()


async def _google_start_resumable_upload(path: GooglePath) -> str:
    request = await googlify_request(
        Request(
            method="POST",
            url=path.format_url(
                "https://storage.googleapis.com/upload/storage/v1/b/{bucket}/o?uploadType=resumable"
            ),
            data=dict(name=path.blob),
            headers={"Content-Type": "application/json; charset=UTF-8"},
            failure_exceptions={400: FileNotFoundError(), 404: FileNotFoundError()},
        )
    )
    async with request.execute() as resp:
        upload_url = resp.headers["Location"]
        return upload_url


async def _google_finalise_upload(upload_url: str, total_size: int) -> None:
    headers = {"Content-Type": "application/octet-stream", "Content-Range": f"bytes */{total_size}"}
    request = await googlify_request(
        Request(method="PUT", url=upload_url, headers=headers, success_codes=(200, 201))
    )
    await request.execute_reponseless()



from __future__ import annotations

import base64
import datetime
import hmac
import json
import os
import re
import time
import urllib.parse
from typing import TYPE_CHECKING, Any, Dict, List, Mapping, Optional, Sequence, Tuple

import xmltodict

if TYPE_CHECKING:
    from .path import AzurePath
    from .request import Request


AZURE_SAS_TOKEN_EXPIRATION_SECONDS = 60 * 60
# these seem to be expired manually, but we don't currently detect that
AZURE_SHARED_KEY_EXPIRATION_SECONDS = 24 * 60 * 60
OAUTH_TOKEN = "oauth_token"
SHARED_KEY = "shared_key"


def load_credentials() -> Dict[str, Any]:
    # AZURE_STORAGE_KEY seems to be the environment variable mentioned by the az cli
    # AZURE_STORAGE_ACCOUNT_KEY is mentioned elsewhere on the internet
    for varname in ["AZURE_STORAGE_KEY", "AZURE_STORAGE_ACCOUNT_KEY"]:
        if varname in os.environ:
            result = dict(storageAccountKey=os.environ[varname])
            if "AZURE_STORAGE_ACCOUNT" in os.environ:
                result["account"] = os.environ["AZURE_STORAGE_ACCOUNT"]
            return result

    if "AZURE_APPLICATION_CREDENTIALS" in os.environ:
        creds_path = os.environ["AZURE_APPLICATION_CREDENTIALS"]
        if not os.path.exists(creds_path):
            raise RuntimeError(
                f"Credentials not found at '{creds_path}' specified by environment variable "
                "'AZURE_APPLICATION_CREDENTIALS'"
            )
        with open(creds_path) as f:
            return json.load(f)

    if "AZURE_CLIENT_ID" in os.environ:
        return dict(
            appId=os.environ["AZURE_CLIENT_ID"],
            password=os.environ["AZURE_CLIENT_SECRET"],
            tenant=os.environ["AZURE_TENANT_ID"],
        )

    if "AZURE_STORAGE_CONNECTION_STRING" in os.environ:
        connection_data = {}
        # technically this should be parsed according to the rules in
        # https://www.connectionstrings.com/formating-rules-for-connection-strings/
        for part in os.environ["AZURE_STORAGE_CONNECTION_STRING"].split(";"):
            key, _, val = part.partition("=")
            connection_data[key.lower()] = val
        return dict(
            account=connection_data["accountname"], storageAccountKey=connection_data["accountkey"]
        )

    # look for a refresh token in the az command line credentials
    # https://mikhail.io/2019/07/how-azure-cli-manages-access-tokens/
    default_creds_path = os.path.expanduser("~/.azure/accessTokens.json")
    if os.path.exists(default_creds_path):
        with open(default_creds_path) as f:
            tokens = json.load(f)
            best_token = None
            for token in tokens:
                if best_token is None:
                    best_token = token
                else:
                    # expiresOn may be missing for tokens from service principals
                    if token.get("expiresOn", "") > best_token.get("expiresOn", ""):
                        best_token = token
            if best_token is not None:
                return best_token

    raise RuntimeError(
        """Azure credentials not found, please do one of the following:

1) Log in with 'az login', blobfile will use your default credentials to lookup your storage
   account key
2) Set the environment variable 'AZURE_STORAGE_KEY' to your storage account key which you can
   find by following this guide:
   https://docs.microsoft.com/en-us/azure/storage/common/storage-account-keys-manage
3) Create an account with 'az ad sp create-for-rbac --name <name>' and set the
   'AZURE_APPLICATION_CREDENTIALS' environment variable to the path of the output from that
   command or individually set the 'AZURE_CLIENT_ID', 'AZURE_CLIENT_SECRET', and
   'AZURE_TENANT_ID' environment variables
"""
    )


def load_stored_subscription_ids() -> List[str]:
    """Return a list of subscription ids from the local azure profile.

    The default subscription will appear first in the list.

    """
    default_profile_path = os.path.expanduser("~/.azure/azureProfile.json")
    if not os.path.exists(default_profile_path):
        return []

    with open(default_profile_path, "rb") as f:
        # this file has a UTF-8 BOM
        profile = json.loads(f.read().decode("utf-8-sig"))

    subscriptions = profile["subscriptions"]
    subscriptions.sort(key=lambda x: x["isDefault"], reverse=True)
    return [sub["id"] for sub in subscriptions]


async def get_access_token(cache_key: Tuple[str, Optional[str]]) -> Tuple[Any, float]:
    account, container = cache_key

    now = time.time()
    creds = load_credentials()

    if "storageAccountKey" in creds:
        if "account" in creds:
            if creds["account"] != account:
                raise RuntimeError(
                    f"Found credentials for account '{creds['account']}' but needed credentials "
                    f"for account '{account}'"
                )
        auth = (SHARED_KEY, creds["storageAccountKey"])
        if await can_access_account(account, container, auth):
            return (auth, now + AZURE_SHARED_KEY_EXPIRATION_SECONDS)
        raise RuntimeError(
            f"Found storage account key, but it was unable to access storage account: '{account}'"
        )

    if "accessToken" in creds and creds.get("tokenType") == "Bearer" and "expiresOn" in creds:
        expiration_time = datetime.datetime.fromisoformat(creds["expiresOn"]).timestamp()
        if expiration_time > now:
            auth = (OAUTH_TOKEN, creds["accessToken"])
            if await can_access_account(account, container, auth):
                return (auth, expiration_time)

    if "refreshToken" in creds:
        # we have a refresh token, convert it into an access token for this account
        req = create_access_token_request(
            creds=creds, scope=f"https://{account}.blob.core.windows.net/", success_codes=(200, 400)
        )

        async with req.execute() as resp:
            result = await resp.json()
            if resp.status == 400:
                if (
                    result["error"] == "invalid_grant"
                    and "AADSTS700082" in result["error_description"]
                ) or (
                    result["error"] == "interaction_required"
                    and "AADSTS50078" in result["error_description"]
                ):
                    raise RuntimeError(
                        "Your refresh token has expired, please run `az login` to refresh it"
                    )
                raise RuntimeError(
                    "Encountered an error when requesting an access token:\n"
                    f"{result['error']}: {result['error_description']}\n\n"
                    "Try running `az login`?"
                )

        auth = (OAUTH_TOKEN, result["access_token"])

        # for some azure accounts this access token does not work, check if it works
        if await can_access_account(account, container, auth):
            return (auth, now + float(result["expires_in"]))

        # it didn't work, fall back to getting the storage keys
        storage_account_key_auth = await get_storage_account_key(
            account=account, creds=creds, container_hint=container
        )
        if storage_account_key_auth is not None:
            return (storage_account_key_auth, now + AZURE_SHARED_KEY_EXPIRATION_SECONDS)

    if "appId" in creds:
        # we have a service principal, get an oauth token
        req = create_access_token_request(creds=creds, scope="https://storage.azure.com/")

        async with req.execute() as resp:
            result = await resp.json()
        auth = (OAUTH_TOKEN, result["access_token"])
        if await can_access_account(account, container, auth):
            return (auth, now + float(result["expires_in"]))

        # it didn't work, fall back to getting the storage keys
        storage_account_key_auth = await get_storage_account_key(
            account=account, creds=creds, container_hint=container
        )
        if storage_account_key_auth is not None:
            return (storage_account_key_auth, now + AZURE_SHARED_KEY_EXPIRATION_SECONDS)

    raise RuntimeError(
        f"Could not find any credentials that grant access to storage account: '{account}'"
    )


async def can_access_account(account: str, container: Optional[str], auth: Tuple[str, str]) -> bool:
    from .request import Request, azurify_request

    if not container:
        # if a container isn't specified, check that we can list the storage account
        req = Request(
            method="GET",
            url=f"https://{account}.blob.core.windows.net",
            params={"comp": "list", "maxresults": "1"},
            success_codes=(200, 403),
        )
        req = await azurify_request(req, auth=auth)

        async with req.execute() as resp:
            if resp.status == 403:
                return False
            data = await resp.read()

        out = xmltodict.parse(data)
        if out["EnumerationResults"]["Containers"] is None:
            # there are no containers in this storage account
            # we can't test if we can access this storage account or not, so presume we can
            return True

        # then also test that we can list a container. this is perhaps unnecessary...
        container = out["EnumerationResults"]["Containers"]["Container"]["Name"]

    # https://myaccount.blob.core.windows.net/mycontainer?restype=container&comp=list
    req = Request(
        method="GET",
        url=f"https://{account}.blob.core.windows.net/{container}",
        params={"restype": "container", "comp": "list", "maxresults": "1"},
        success_codes=(200, 403),
    )
    req = await azurify_request(req, auth=auth)
    async with req.execute() as resp:
        return resp.status == 200


def create_access_token_request(
    creds: Mapping[str, str], scope: str, success_codes: Sequence[int] = (200,)
) -> Request:
    from .request import Request

    if "refreshToken" in creds:
        # https://docs.microsoft.com/en-us/azure/active-directory/develop/v1-protocols-oauth-code#refreshing-the-access-tokens
        data = {
            "grant_type": "refresh_token",
            "refresh_token": creds["refreshToken"],
            "resource": scope,
        }
        tenant = "common"
    else:
        # https://docs.microsoft.com/en-us/azure/active-directory/develop/v1-oauth2-client-creds-grant-flow#request-an-access-token
        # https://docs.microsoft.com/en-us/azure/active-directory/develop/v1-protocols-oauth-code
        # https://docs.microsoft.com/en-us/rest/api/storageservices/authorize-with-azure-active-directory#use-oauth-access-tokens-for-authentication
        # https://docs.microsoft.com/en-us/rest/api/azure/
        # https://docs.microsoft.com/en-us/rest/api/storageservices/authorize-with-azure-active-directory
        # az ad sp create-for-rbac --name <name>
        # az account list
        # az role assignment create --role "Storage Blob Data Contributor" --assignee <appid> --scope "/subscriptions/<account id>"
        data = {
            "grant_type": "client_credentials",
            "client_id": creds["appId"],
            "client_secret": creds["password"],
            "resource": scope,
        }
        tenant = creds["tenant"]
    return Request(
        url=f"https://login.microsoftonline.com/{tenant}/oauth2/token",
        method="POST",
        headers={"Content-Type": "application/x-www-form-urlencoded"},
        data=urllib.parse.urlencode(data).encode("utf8"),
        success_codes=success_codes,
    )


async def get_storage_account_id_with_subscription(
    subscription_id: str, account: str, auth: Tuple[str, str]
) -> Optional[str]:
    from .request import Request, azurify_request

    req = Request(
        method="GET",
        url=f"https://management.azure.com/subscriptions/{subscription_id}/providers/Microsoft.Storage/storageAccounts",
        params={"api-version": "2019-04-01"},
        success_codes=(200, 401, 403),
    )
    req = await azurify_request(req, auth=auth)

    async with req.execute() as resp:
        if resp.status in (401, 403):
            # we aren't allowed to query this for this subscription, skip it
            return None
        out = await resp.json()

    # search for the storage account
    return next((obj["id"] for obj in out["value"] if obj["name"] == account), None)


async def get_storage_account_id(account: str, auth: Tuple[str, str]) -> Optional[str]:
    from .request import Request, azurify_request

    stored_subscription_ids = load_stored_subscription_ids()
    for subscription_id in stored_subscription_ids:
        storage_account_id = await get_storage_account_id_with_subscription(
            subscription_id, account, auth
        )
        if storage_account_id:
            return storage_account_id

    req = Request(
        method="GET",
        url="https://management.azure.com/subscriptions",
        params={"api-version": "2020-01-01"},
    )
    req = await azurify_request(req, auth=auth)
    async with req.execute() as resp:
        result = await resp.json()
    subscription_ids = {item["subscriptionId"] for item in result["value"]}

    for subscription_id in subscription_ids - set(stored_subscription_ids):
        storage_account_id = await get_storage_account_id_with_subscription(
            subscription_id, account, auth
        )
        if storage_account_id:
            return storage_account_id
    return None


async def get_storage_account_key(
    account: str, creds: Mapping[str, Any], container_hint: Optional[str] = None
) -> Optional[Tuple[Any, float]]:
    from .request import Request, azurify_request

    # get an access token for the management service
    req = create_access_token_request(creds=creds, scope="https://management.azure.com/")

    async with req.execute() as resp:
        result = await resp.json()
    auth = (OAUTH_TOKEN, result["access_token"])

    storage_account_id = await get_storage_account_id(account, auth)
    if not storage_account_id:
        return None

    req = Request(
        method="POST",
        url=f"https://management.azure.com{storage_account_id}/listKeys",
        params={"api-version": "2019-04-01"},
    )
    req = await azurify_request(req, auth=auth)

    async with req.execute() as resp:
        result = await resp.json()
    for key in result["keys"]:
        if key["permissions"] == "FULL":
            storage_key_auth = (SHARED_KEY, key["value"])
            if await can_access_account(account, container_hint, storage_key_auth):
                return storage_key_auth
            raise RuntimeError(
                f"Found storage account key, but it was unable to access storage account: '{account}'"
            )
    raise RuntimeError(
        f"Storage account was found, but storage account keys were missing: '{account}'"
    )


def sign_request_with_shared_key(request: Request, key: str) -> str:
    # https://docs.microsoft.com/en-us/rest/api/storageservices/authorize-with-shared-key
    from yarl import URL

    # Figuring out that shared key authorisation sometimes breaks because aiohttp is a little over
    # clever about canonicalising URLs was not fun; preemptively canonicalise
    parsed_url = urllib.parse.urlparse(str(URL(request.url)))
    storage_account = parsed_url.netloc.split(".")[0]
    blob = parsed_url.path[1:]

    def canonicalized_resource() -> str:
        params_to_sign = sorted(f"{name.lower()}:{value}" for name, value in request.params.items())
        canonical_url = f"/{storage_account}/{blob}"
        return "\n".join([canonical_url] + params_to_sign)

    def canonicalized_headers() -> str:
        headers_to_sign = []
        for name, value in request.headers.items():
            canonical_name = name.lower()
            canonical_value = re.sub(r"\s+", " ", value).strip()
            if canonical_name.startswith("x-ms-"):
                headers_to_sign.append(f"{canonical_name}:{canonical_value}")
        return "\n".join(sorted(headers_to_sign))

    headers = request.headers
    content_length = headers.get("Content-Length", "")
    if request.data is not None:
        content_length = str(len(request.data))

    parts_to_sign = [
        request.method,
        headers.get("Content-Encoding", ""),
        headers.get("Content-Language", ""),
        content_length,
        headers.get("Content-MD5", ""),
        headers.get("Content-Type", ""),
        headers.get("Date", ""),
        headers.get("If-Modified-Since", ""),
        headers.get("If-Match", ""),
        headers.get("If-None-Match", ""),
        headers.get("If-Unmodified-Since", ""),
        headers.get("Range", ""),
        canonicalized_headers(),
        canonicalized_resource(),
    ]
    string_to_sign = "\n".join(parts_to_sign)

    signature = base64.b64encode(
        hmac.digest(base64.b64decode(key), string_to_sign.encode("utf8"), "sha256")
    ).decode("utf8")

    return f"SharedKey {storage_account}:{signature}"


async def get_sas_token(cache_key: Tuple[str, Optional[str]]) -> Tuple[Any, float]:
    from .globals import config
    from .request import azurify_request

    auth = await config.azure_access_token_manager.get_token(key=cache_key)
    account, container = cache_key

    if auth[0] != OAUTH_TOKEN:
        cmd = (
            f"az storage container list --auth-mode login --account-name {account}"
            if container is None
            else f"az storage blob list --auth-mode login --account-name {account} --container {container}"
        )
        raise RuntimeError(
            "Only OAuth tokens can be used to get SAS tokens. You should set the "
            "Blob Data Reader or Storage Blob Data Contributor IAM role. You can run "
            f"`{cmd}` to confirm that the missing role is the issue."
        )

    req = create_user_delegation_sas_request(account=account)
    req = await azurify_request(req, auth=auth)

    async with req.execute() as resp:
        data = await resp.read()

    out = xmltodict.parse(data)
    return out["UserDelegationKey"], time.time() + AZURE_SAS_TOKEN_EXPIRATION_SECONDS


def create_user_delegation_sas_request(account: str) -> Request:
    # https://docs.microsoft.com/en-us/rest/api/storageservices/create-user-delegation-sas
    from .request import Request

    now = datetime.datetime.utcnow()
    start = (now + datetime.timedelta(hours=-1)).strftime("%Y-%m-%dT%H:%M:%SZ")
    expiration = now + datetime.timedelta(days=6)
    expiry = expiration.strftime("%Y-%m-%dT%H:%M:%SZ")
    return Request(
        url=f"https://{account}.blob.core.windows.net/",
        method="POST",
        params=dict(restype="service", comp="userdelegationkey"),
        data={"KeyInfo": {"Start": start, "Expiry": expiry}},
    )


async def generate_signed_url(path: AzurePath) -> Tuple[str, datetime.datetime]:
    # https://docs.microsoft.com/en-us/rest/api/storageservices/delegate-access-with-shared-access-signature
    # https://docs.microsoft.com/en-us/rest/api/storageservices/create-user-delegation-sas
    # https://docs.microsoft.com/en-us/rest/api/storageservices/service-sas-examples
    from .globals import config

    key = await config.azure_sas_token_manager.get_token(key=(path.account, path.container or None))
    params = {
        "st": key["SignedStart"],
        "se": key["SignedExpiry"],
        "sks": key["SignedService"],
        "skt": key["SignedStart"],
        "ske": key["SignedExpiry"],
        "sktid": key["SignedTid"],
        "skoid": key["SignedOid"],
        # signed key version (param name not mentioned in docs)
        "skv": key["SignedVersion"],
        "sv": "2018-11-09",  # signed version
        "sr": "b",  # signed resource
        "sp": "r",  # signed permissions
        "sip": "",  # signed ip
        "si": "",  # signed identifier
        "spr": "https,http",  # signed http protocol
        "rscc": "",  # Cache-Control header
        "rscd": "",  # Content-Disposition header
        "rsce": "",  # Content-Encoding header
        "rscl": "",  # Content-Language header
        "rsct": "",  # Content-Type header
    }

    canonicalized_resource = f"/blob/{path.account}/{path.container}/{path.blob}"
    parts_to_sign = (
        params["sp"],
        params["st"],
        params["se"],
        canonicalized_resource,
        params["skoid"],
        params["sktid"],
        params["skt"],
        params["ske"],
        params["sks"],
        params["skv"],
        params["sip"],
        params["spr"],
        params["sv"],
        params["sr"],
        params["rscc"],
        params["rscd"],
        params["rsce"],
        params["rscl"],
        params["rsct"],
        # this is documented on a different page
        # https://docs.microsoft.com/en-us/rest/api/storageservices/create-service-sas#specifying-the-signed-identifier
        params["si"],
    )
    string_to_sign = "\n".join(parts_to_sign)
    params["sig"] = base64.b64encode(
        hmac.digest(base64.b64decode(key["Value"]), string_to_sign.encode("utf8"), "sha256")
    ).decode("utf8")
    query = urllib.parse.urlencode({k: v for k, v in params.items() if v != ""})

    expiry = datetime.datetime.strptime(key["SignedExpiry"], "%Y-%m-%dT%H:%M:%SZ")
    url = path.format_url("https://{account}.blob.core.windows.net/{container}/{blob}")
    return url + "?" + query, expiry



from __future__ import annotations

import itertools
from typing import Any, Iterator, Optional, Tuple, Union

from .boost import BoostExecutor, BoostUnderlying, OrderedBoostable, UnorderedBoostable
from .globals import config
from .path import AzurePath, BasePath, CloudPath, GooglePath, LocalPath, getsize, pathdispatch
from .request import Request, azurify_request, googlify_request

ByteRange = Tuple[int, int]
OptByteRange = Tuple[Optional[int], Optional[int]]

# ==============================
# read_byte_range
# ==============================


@pathdispatch
async def read_byte_range(path: Union[CloudPath, str], byte_range: OptByteRange) -> bytes:
    """Read the content of ``path`` in the given byte range.

    :param path: The path to read from.
    :param byte_range: The byte range to read.
    :return: The relevant bytes.

    """
    raise ValueError(f"Unsupported path: {path}")


@read_byte_range.register  # type: ignore
async def _azure_read_byte_range(path: AzurePath, byte_range: OptByteRange) -> bytes:
    range_str = _byte_range_to_str(byte_range)
    range_header = {"Range": range_str} if range_str is not None else {}
    success_codes = (206,) if range_header else (200,)
    request = await azurify_request(
        Request(
            method="GET",
            url=path.format_url("https://{account}.blob.core.windows.net/{container}/{blob}"),
            headers=range_header,
            success_codes=success_codes,
        )
    )
    async with request.execute() as resp:
        return await resp.read()


@read_byte_range.register  # type: ignore
async def _google_read_byte_range(path: GooglePath, byte_range: OptByteRange) -> bytes:
    range_str = _byte_range_to_str(byte_range)
    range_header = {"Range": range_str} if range_str is not None else {}
    success_codes = (206,) if range_header else (200,)
    request = await googlify_request(
        Request(
            method="GET",
            url=path.format_url("https://storage.googleapis.com/storage/v1/b/{bucket}/o/{blob}"),
            params=dict(alt="media"),
            headers=range_header,
            success_codes=success_codes,
        )
    )
    async with request.execute() as resp:
        return await resp.read()


# ==============================
# read_single
# ==============================


@pathdispatch
async def read_single(path: Union[BasePath, str]) -> bytes:
    """Read the content of ``path``.

    :param path: The path to read from.

    """
    raise ValueError(f"Unsupported path: {path}")


@read_single.register  # type: ignore
async def _cloud_read_single(path: CloudPath) -> bytes:
    return await read_byte_range(path, (None, None))


@read_single.register  # type: ignore
async def _local_read_single(path: LocalPath) -> bytes:
    with open(path, "rb") as f:
        return f.read()


# ==============================
# read_stream
# ==============================


@pathdispatch
async def read_stream(
    path: Union[BasePath, str], executor: BoostExecutor, size: Optional[int] = None
) -> BoostUnderlying[bytes]:
    """Read the content of ``path``.

    :param path: The path to read from.
    :param executor: An executor.
    :param size: If specified, will save a network call.
    :return: The stream of bytes, chunking determined by ``config.chunk_size``.

    """
    raise ValueError(f"Unsupported path: {path}")


@read_stream.register  # type: ignore
async def _cloud_read_stream(
    path: CloudPath, executor: BoostExecutor, size: Optional[int] = None
) -> OrderedBoostable[Any, bytes]:
    if size is None:
        size = await getsize(path)

    byte_ranges = itertools.zip_longest(
        range(0, size, config.chunk_size),
        range(config.chunk_size, size, config.chunk_size),
        fillvalue=size,
    )

    # Note that we purposefully don't do
    # https://docs.aiohttp.org/en/stable/client_quickstart.html#streaming-response-content
    # Doing that would stream data as we needed it, which is a little too lazy for our purposes
    chunks = executor.map_ordered(lambda byte_range: read_byte_range(path, byte_range), byte_ranges)
    return chunks


@read_stream.register  # type: ignore
async def _local_read_stream(
    path: LocalPath, executor: BoostExecutor, size: Optional[int] = None
) -> Iterator[bytes]:
    def iterator() -> Iterator[bytes]:
        with open(path, "rb") as f:
            while True:
                data = f.read(config.chunk_size)
                if not data:
                    return
                yield data

    return iterator()


# ==============================
# read_stream_unordered
# ==============================


@pathdispatch
async def read_stream_unordered(
    path: Union[CloudPath, str], executor: BoostExecutor, size: Optional[int] = None
) -> UnorderedBoostable[Any, Tuple[bytes, ByteRange]]:
    assert isinstance(path, CloudPath)

    if size is None:
        size = await getsize(path)

    byte_ranges = itertools.zip_longest(
        range(0, size, config.chunk_size),
        range(config.chunk_size, size, config.chunk_size),
        fillvalue=size,
    )

    async def read_byte_range_wrapper(byte_range: ByteRange) -> Tuple[bytes, ByteRange]:
        chunk = await read_byte_range(path, byte_range)
        return (chunk, byte_range)

    chunks = executor.map_unordered(read_byte_range_wrapper, byte_ranges)
    return chunks


# ==============================
# helpers
# ==============================


def _byte_range_to_str(byte_range: OptByteRange) -> Optional[str]:
    # https://docs.microsoft.com/en-us/rest/api/storageservices/specifying-the-range-header-for-blob-service-operations
    # https://cloud.google.com/storage/docs/xml-api/get-object-download
    # oddly range requests are not mentioned in JSON API, only in the XML API
    start, end = byte_range
    if start is not None and end is not None:
        return f"bytes={start}-{end-1}"
    if start is not None:
        return f"bytes={start}-"
    if end is not None:
        if end > 0:
            return f"bytes=0-{end-1}"
        # This form is not supported by Azure
        return f"bytes=-{-int(end)}"
    return None



import asyncio
import contextlib
import datetime
import json
import random
import socket
import sys
import time
import urllib.parse
from dataclasses import dataclass, field
from typing import Any, AsyncIterator, Dict, Iterator, Mapping, Optional, Sequence, Tuple

import aiohttp
import xmltodict

from .globals import config


class MissingSession(Exception):
    pass


@dataclass
class Request:
    method: str
    url: str
    params: Mapping[str, str] = field(default_factory=dict)
    # data can be large, so don't put it in the repr
    data: Any = field(repr=False, default_factory=lambda: None)
    # headers can contain authorisation details, so don't put them in the repr
    headers: Mapping[str, str] = field(repr=False, default_factory=dict)
    success_codes: Sequence[int] = field(repr=False, default_factory=lambda: (200,))
    retry_codes: Sequence[int] = field(
        repr=False, default_factory=lambda: (408, 429, 500, 502, 503, 504)
    )
    failure_exceptions: Mapping[int, Exception] = field(repr=False, default_factory=dict)

    @contextlib.asynccontextmanager
    async def execute(self) -> AsyncIterator[aiohttp.ClientResponse]:
        """Execute the request, retrying as appropriate.

        Example usage:
        ```
        async with request.execute() as resp:
            print(resp.status)
        ```

        """

        for attempt, backoff in enumerate(
            exponential_sleep_generator(
                initial=config.backoff_initial,
                maximum=config.backoff_max,
                jitter_fraction=config.backoff_jitter_fraction,
            )
        ):
            try:
                async with self._raw_execute() as resp:
                    if resp.status in self.success_codes:
                        yield resp
                        return

                    assert resp.reason is not None
                    error = RequestFailure(
                        reason=str(resp.reason), request=self, status=resp.status
                    )
                    if resp.status not in self.retry_codes:
                        raise self.failure_exceptions.get(resp.status, error)

            except aiohttp.ClientConnectionError as e:
                if isinstance(e, aiohttp.ClientConnectorError):
                    # azure accounts have unique urls and it's hard to tell apart
                    # an invalid hostname from a network error
                    url = urllib.parse.urlparse(self.url)
                    hostname = url.hostname
                    if hostname and hostname.endswith(".blob.core.windows.net"):
                        if await _bad_hostname_check(hostname):
                            raise FileNotFoundError(hostname) from None
                error = RequestFailure(reason=type(e).__name__ + ": " + str(e), request=self)

            if attempt >= config.retry_limit:
                raise error

            if config.debug_mode or attempt + 1 >= 3:
                print(
                    f"[boostedblob] Error when executing request on attempt {attempt + 1}, sleeping for "
                    f"{backoff:.1f}s before retrying. Details: {error}",
                    file=sys.stderr,
                )
            await asyncio.sleep(backoff)

    async def execute_reponseless(self) -> None:
        """Helper to execute the request when we don't have further need of the response."""
        async with self.execute():
            pass

    @contextlib.asynccontextmanager
    async def _raw_execute(self) -> AsyncIterator[aiohttp.ClientResponse]:
        """Actually execute the request, with no extra fluff."""
        # TODO: investigate asyncio unclosed transport warnings from
        # -X dev -X tracemalloc=20
        if config.session is None:
            raise MissingSession(
                "No session available, use `async with session_context()` or `@ensure_session`"
            )
        ctx = config.session.request(
            method=self.method,
            url=self.url,
            params=self.params,
            data=self.data,
            headers=self.headers,
            allow_redirects=False,
            timeout=aiohttp.ClientTimeout(
                connect=config.connect_timeout, sock_read=config.read_timeout
            ),
            # Figuring out that some requests break because aiohttp adds some headers
            # automatically was not fun
            skip_auto_headers={"Content-Type"},
        )
        if config.debug_mode:
            print(f"[boostedblob] Making request: {self}", file=sys.stderr)
            now = time.time()
        async with ctx as resp:
            if config.debug_mode:
                duration = time.time() - now
                print(
                    f"[boostedblob] Completed request, took {duration:.3f}s: {self}",
                    file=sys.stderr,
                )
            yield resp


class RequestFailure(Exception):
    def __init__(self, reason: str, request: Request, status: Optional[int] = None):
        self.reason = reason
        self.request = request
        self.status = status

    def __str__(self) -> str:
        return f"{self.reason}, {self.status}, {self.request}"


# ==============================
# cloud specific utilities
# ==============================


async def azurify_request(request: Request, auth: Optional[Tuple[str, str]] = None) -> Request:
    """Return a Request that can be submitted to Azure Blob."""
    u = urllib.parse.urlparse(request.url)
    account = u.netloc.split(".")[0]
    parts = u.path.split("/", maxsplit=2)
    container = parts[1] if len(parts) >= 2 else None
    if auth is None:
        auth = await config.azure_access_token_manager.get_token(key=(account, container))
    assert auth is not None
    kind, token = auth

    headers = dict(request.headers)
    # https://docs.microsoft.com/en-us/rest/api/storageservices/previous-azure-storage-service-versions
    headers["x-ms-version"] = "2019-12-12"
    headers["x-ms-date"] = datetime.datetime.utcnow().strftime("%a, %d %b %Y %H:%M:%S GMT")

    from .azure_auth import OAUTH_TOKEN, SHARED_KEY, sign_request_with_shared_key

    data = request.data
    if data is not None and not isinstance(data, (bytes, bytearray)):
        data = xmltodict.unparse(data).encode("utf8")

    result = Request(
        method=request.method,
        url=request.url,
        params=request.params,
        headers=headers,
        data=data,
        success_codes=request.success_codes,
        retry_codes=request.retry_codes,
        failure_exceptions=request.failure_exceptions,
    )

    # mutate headers to add the authorization header to result
    if kind == SHARED_KEY:
        # make sure we are signing a request that includes changes made in this function
        # i.e., the request has the x-ms-* headers added and data bytes-ified.
        assert "x-ms-date" in headers
        headers["Authorization"] = sign_request_with_shared_key(result, token)
    elif kind == OAUTH_TOKEN:
        headers["Authorization"] = f"Bearer {token}"

    return result


async def googlify_request(request: Request, access_token: Optional[str] = None) -> Request:
    """Return a Request that can be submitted to GCS."""
    if access_token is None:
        access_token = await config.google_access_token_manager.get_token(key="")

    headers = dict(request.headers)
    headers["Authorization"] = f"Bearer {access_token}"

    data = request.data
    if data is not None and not isinstance(data, (bytes, bytearray)):
        data = json.dumps(data).encode("utf8")

    return Request(
        method=request.method,
        url=request.url,
        params=request.params,
        headers=headers,
        data=data,
        success_codes=request.success_codes,
        retry_codes=request.retry_codes,
        failure_exceptions=request.failure_exceptions,
    )


async def azure_page_iterator(request: Request) -> AsyncIterator[Dict[str, Any]]:
    params = dict(request.params)
    while True:
        request = Request(
            method=request.method,
            url=request.url,
            params=params,
            data=request.data,
            headers=request.headers,
            success_codes=(200, 404),
            retry_codes=request.retry_codes,
            failure_exceptions=request.failure_exceptions,
        )
        request = await azurify_request(request)
        async with request.execute() as resp:
            if resp.status == 404:
                return
            body = await resp.read()
            result = xmltodict.parse(body)["EnumerationResults"]
        yield result
        if result["NextMarker"] is None:
            break
        params["marker"] = result["NextMarker"]


async def google_page_iterator(request: Request) -> AsyncIterator[Dict[str, Any]]:
    params = dict(request.params)
    while True:
        request = Request(
            method=request.method,
            url=request.url,
            params=params,
            headers=request.headers,
            success_codes=(200, 404),
            retry_codes=request.retry_codes,
            failure_exceptions=request.failure_exceptions,
        )
        request = await googlify_request(request)
        async with request.execute() as resp:
            if resp.status == 404:
                return
            result = await resp.json()
        yield result
        if "nextPageToken" not in result:
            break
        params["pageToken"] = result["nextPageToken"]


# ==============================
# helpers
# ==============================


def exponential_sleep_generator(
    initial: float, maximum: float, jitter_fraction: float, multiplier: float = 2
) -> Iterator[float]:
    """Yields amounts to sleep by.

    https://aws.amazon.com/blogs/architecture/exponential-backoff-and-jitter/
    Full jitter is ``jitter_fraction == 1``
    Equal jitter is ``jitter_fraction == 0.5``
    No jitter is ``jitter_fraction == 0``
    """

    base = initial
    while True:
        sleep = base * (1 - jitter_fraction) + base * random.random() * jitter_fraction
        yield sleep
        base *= multiplier
        if base > maximum:
            base = maximum


async def _bad_hostname_check(hostname: str) -> bool:
    """Return True if the hostname likely does not exist.

    If the hostname exists, or there is uncertainty, return False.

    """
    loop = asyncio.get_event_loop()
    try:
        await loop.getaddrinfo(hostname, None, family=socket.AF_INET)
        # no errors encountered, the hostname exists
        return False
    except socket.gaierror as e:
        if e.errno != socket.EAI_NONAME:
            # we got some sort of other socket error, so it's unclear if the host exists or not
            return False
        if sys.platform == "linux":
            # on linux we appear to get EAI_NONAME if the host does not exist and EAI_AGAIN if
            # there is a temporary failure in resolution
            return True
        # it's not clear on other platforms how to differentiate a temporary name resolution
        # failure from a permanent one, EAI_NONAME seems to be returned for either case if we
        # cannot look up the hostname, but we can look up google, then it's likely the hostname
        # does not exist
        try:
            await loop.getaddrinfo("www.google.com", None, family=socket.AF_INET)
        except socket.gaierror:
            # if we can't resolve google, then the network is likely down and we don't know if
            # the hostname exists or not
            return False
        # in this case, we could resolve google, but not the original hostname likely the
        # hostname does not exist (though this is definitely not a foolproof check)
        return True



from __future__ import annotations

import asyncio
import base64
import datetime
import functools
import os
import urllib.parse
from dataclasses import dataclass
from typing import Any, Callable, Mapping, NamedTuple, Optional, TypeVar, Union

from .request import Request, azure_page_iterator, azurify_request, googlify_request

T = TypeVar("T")

# ==============================
# Path classes
# ==============================


class BasePath:
    @staticmethod
    def from_str(path: str) -> BasePath:
        url = urllib.parse.urlparse(path)
        if url.scheme == "gs":
            return GooglePath.from_str(path)
        if url.scheme == "az" or (
            url.scheme == "https" and url.netloc.endswith(".blob.core.windows.net")
        ):
            return AzurePath.from_str(path)
        if url.scheme:
            raise ValueError(f"Invalid path '{path}'")
        return LocalPath(path)

    @property
    def name(self) -> str:
        """Returns the name of path, normalised to exclude any trailing slash."""
        raise NotImplementedError

    @property
    def parent(self: T) -> T:
        raise NotImplementedError

    def relative_to(self: T, other: T) -> str:
        raise NotImplementedError

    def is_relative_to(self: T, other: T) -> bool:
        try:
            self.relative_to(other)  # type: ignore
            return True
        except ValueError:
            return False

    def is_directory_like(self) -> bool:
        raise NotImplementedError

    def ensure_directory_like(self: T) -> T:
        raise NotImplementedError

    def __truediv__(self: T, relative_path: str) -> T:
        raise NotImplementedError


@dataclass(frozen=True)
class LocalPath(BasePath):
    path: str

    @staticmethod
    def from_str(path: str) -> BasePath:
        url = urllib.parse.urlparse(path)
        if url.scheme:
            raise ValueError(f"Invalid path '{path}'")
        return LocalPath(path)

    @property
    def name(self) -> str:
        return os.path.basename(_strip_slash(self.path))

    @property
    def parent(self) -> LocalPath:
        return LocalPath(os.path.dirname(self.path) or ".")

    def relative_to(self, other: LocalPath) -> str:
        if not isinstance(other, LocalPath):
            raise ValueError(f"'{other}' is not a subpath of '{self}'")
        other = other.ensure_directory_like()
        other_path = other.path
        if other_path.startswith("./") and not self.path.startswith("./"):
            other_path = other_path[len("./") :]
        if not self.path.startswith(other_path):
            raise ValueError(f"'{other}' is not a subpath of '{self}'")
        return self.path[len(other_path) :]

    def is_directory_like(self) -> bool:
        return not self.path or self.path.endswith("/")

    def ensure_directory_like(self) -> LocalPath:
        return self if self.is_directory_like() else LocalPath(self.path + "/")

    def __truediv__(self, relative_path: str) -> LocalPath:
        return LocalPath(os.path.join(self.path, relative_path))

    def __str__(self) -> str:
        return self.path

    def __fspath__(self) -> str:
        return self.path


class CloudPath(BasePath):
    pass


@dataclass(frozen=True)
class AzurePath(CloudPath):
    account: str
    container: str
    blob: str

    @staticmethod
    def from_str(url: str) -> AzurePath:
        parsed_url = urllib.parse.urlparse(url)
        if parsed_url.scheme == "https":
            account, host = parsed_url.netloc.split(".", maxsplit=1)
            if host != "blob.core.windows.net":
                raise ValueError(f"Invalid URL '{url}'")
        elif parsed_url.scheme == "az":
            account = parsed_url.netloc
        else:
            raise ValueError(f"Invalid URL '{url}'")

        parts = parsed_url.path.split("/", maxsplit=2)
        if parts[0]:
            raise ValueError(f"Invalid URL '{url}'")
        container = parts[1] if len(parts) >= 2 else ""
        return AzurePath(account=account, container=container, blob="/".join(parts[2:]))

    @property
    def name(self) -> str:
        return os.path.basename(_strip_slash(self.blob)) if self.blob else self.container

    @property
    def parent(self) -> AzurePath:
        return AzurePath(self.account, self.container, os.path.dirname(self.blob))

    def relative_to(self, other: AzurePath) -> str:
        if not isinstance(other, AzurePath):
            raise ValueError(f"'{other}' is not a subpath of '{self}'")
        other = other.ensure_directory_like()
        if (
            self.account != other.account
            or self.container != other.container
            or not self.blob.startswith(other.blob)
        ):
            raise ValueError(f"'{other}' is not a subpath of '{self}'")
        return self.blob[len(other.blob) :]

    def is_directory_like(self) -> bool:
        return not self.blob or self.blob.endswith("/")

    def ensure_directory_like(self) -> AzurePath:
        return (
            self
            if self.is_directory_like()
            else AzurePath(self.account, self.container, self.blob + "/")
        )

    def __truediv__(self, relative_path: str) -> AzurePath:
        return AzurePath(self.account, self.container, os.path.join(self.blob, relative_path))

    def __str__(self) -> str:
        return f"https://{self.account}.blob.core.windows.net/{self.container}/{self.blob}"

    def to_az_str(self) -> str:
        return f"az://{self.account}/{self.container}/{self.blob}"

    def format_url(self, template: str) -> str:
        return url_format(template, account=self.account, container=self.container, blob=self.blob)


@dataclass(frozen=True)
class GooglePath(CloudPath):
    bucket: str
    blob: str

    @staticmethod
    def from_str(url: str) -> GooglePath:
        parsed_url = urllib.parse.urlparse(url)
        if parsed_url.scheme != "gs":
            raise ValueError(f"Invalid URL '{url}'")

        return GooglePath(bucket=parsed_url.netloc, blob=parsed_url.path[1:])

    @property
    def name(self) -> str:
        return os.path.basename(_strip_slash(self.blob)) if self.blob else self.bucket

    @property
    def parent(self) -> GooglePath:
        return GooglePath(self.bucket, os.path.dirname(self.blob))

    def relative_to(self, other: GooglePath) -> str:
        if not isinstance(other, GooglePath):
            raise ValueError(f"'{other}' is not a subpath of '{self}'")
        other = other.ensure_directory_like()
        if self.bucket != other.bucket or not self.blob.startswith(other.blob):
            raise ValueError(f"'{other}' is not a subpath of '{self}'")
        return self.blob[len(other.blob) :]

    def is_directory_like(self) -> bool:
        return not self.blob or self.blob.endswith("/")

    def ensure_directory_like(self) -> GooglePath:
        return self if self.is_directory_like() else GooglePath(self.bucket, self.blob + "/")

    def __truediv__(self, relative_path: str) -> GooglePath:
        return GooglePath(self.bucket, os.path.join(self.blob, relative_path))

    def __str__(self) -> str:
        return f"gs://{self.bucket}/{self.blob}"

    def format_url(self, template: str) -> str:
        return url_format(template, bucket=self.bucket, blob=self.blob)


# ==============================
# pathdispatch
# ==============================

F = TypeVar("F", bound=Callable[..., Any])


def pathdispatch(fn: F) -> F:
    """Implements the singledispatch pattern for paths.

    The extra work this does over singledispatch is that if the path argument is a str, it's
    automatically converted to the correct BasePath object and re-dispatched.

    Typing this function as we've done allows us to exactly type usage, but unfortunately it
    causes type checkers to complain about "no attribute register"

    """
    ret: F = functools.singledispatch(fn)  # type: ignore

    @ret.register  # type: ignore
    def strdispatch(path: str, *args: Any, **kwargs: Any) -> Any:
        return ret(BasePath.from_str(path), *args, **kwargs)

    return ret


# ==============================
# stat
# ==============================


class Stat(NamedTuple):
    size: int
    mtime: float
    ctime: float
    md5: Optional[str]
    version: Optional[str]

    @staticmethod
    def from_azure(item: Mapping[str, Any]) -> Stat:
        if "Creation-Time" in item:
            raw_ctime = item["Creation-Time"]
        else:
            raw_ctime = item["x-ms-creation-time"]
        if "x-ms-meta-blobfilemtime" in item:
            mtime = float(item["x-ms-meta-blobfilemtime"])
        else:
            mtime = _azure_parse_timestamp(item["Last-Modified"])
        return Stat(
            size=int(item["Content-Length"]),
            mtime=mtime,
            ctime=_azure_parse_timestamp(raw_ctime),
            md5=_azure_get_md5(item),
            version=item["Etag"],
        )

    @staticmethod
    def from_google(item: Mapping[str, Any]) -> Stat:
        if "metadata" in item and "blobfile-mtime" in item["metadata"]:
            mtime = float(item["metadata"]["blobfile-mtime"])
        else:
            mtime = _google_parse_timestamp(item["updated"])
        return Stat(
            size=int(item["size"]),
            mtime=mtime,
            ctime=_google_parse_timestamp(item["timeCreated"]),
            md5=_google_get_md5(item),
            version=item["generation"],
        )

    @staticmethod
    def from_local(item: os.stat_result) -> Stat:
        return Stat(
            size=item.st_size, mtime=item.st_mtime, ctime=item.st_ctime, md5=None, version=None
        )


@pathdispatch
async def stat(path: Union[BasePath, str]) -> Stat:
    raise ValueError(f"Unsupported path: {path}")


@stat.register  # type: ignore
async def _azure_stat(path: AzurePath) -> Stat:
    if not path.blob:
        raise FileNotFoundError(path)
    request = await azurify_request(
        Request(
            method="HEAD",
            url=path.format_url("https://{account}.blob.core.windows.net/{container}/{blob}"),
            failure_exceptions={404: FileNotFoundError(path)},
        )
    )
    async with request.execute() as resp:
        return Stat.from_azure(resp.headers)


@stat.register  # type: ignore
async def _google_stat(path: GooglePath) -> Stat:
    if not path.blob:
        raise FileNotFoundError(path)
    request = await googlify_request(
        Request(
            method="GET",
            url=path.format_url("https://storage.googleapis.com/storage/v1/b/{bucket}/o/{blob}"),
            failure_exceptions={404: FileNotFoundError(path)},
        )
    )
    async with request.execute() as resp:
        result = await resp.json()
        return Stat.from_google(result)


@stat.register  # type: ignore
async def _local_stat(path: LocalPath) -> Stat:
    return Stat.from_local(os.stat(path))


# ==============================
# getsize
# ==============================


@pathdispatch
async def getsize(path: Union[BasePath, str]) -> int:
    s = await stat(path)
    return s.size


# ==============================
# isdir
# ==============================


@pathdispatch
async def isdir(path: Union[BasePath, str], raise_on_missing: bool = False) -> bool:
    """Check whether ``path`` is a directory.

    :param path: The path that could be a directory.
    :param raise_on_missing: If True, raise FileNotFoundError if ``path`` does not exist. Otherwise,
        return False.

    """
    raise ValueError(f"Unsupported path: {path}")


@isdir.register  # type: ignore
async def _azure_isdir(path: AzurePath) -> bool:
    try:
        if path.blob:
            prefix: str = path.blob
            if not prefix.endswith("/"):
                prefix += "/"

            it = azure_page_iterator(
                Request(
                    method="GET",
                    url=path.format_url("https://{account}.blob.core.windows.net/{container}"),
                    params=dict(
                        comp="list",
                        restype="container",
                        prefix=prefix,
                        delimiter="/",
                        maxresults="1",
                    ),
                    failure_exceptions={404: FileNotFoundError()},
                )
            )
            async for result in it:
                if result["Blobs"] is not None:
                    return "BlobPrefix" in result["Blobs"] or "Blob" in result["Blobs"]
            return False
        else:
            request = await azurify_request(
                Request(
                    method="GET",
                    url=path.format_url("https://{account}.blob.core.windows.net/{container}"),
                    params=dict(restype="container"),
                    failure_exceptions={404: FileNotFoundError()},
                )
            )
            await request.execute_reponseless()
            return True
    except FileNotFoundError:
        # execute might raise FileNotFoundError if storage account doesn't exist
        return False


@isdir.register  # type: ignore
async def _google_isdir(path: GooglePath) -> bool:
    try:
        if path.blob:
            prefix: str = path.blob
            if not prefix.endswith("/"):
                prefix += "/"
            request = await googlify_request(
                Request(
                    method="GET",
                    url=path.format_url("https://storage.googleapis.com/storage/v1/b/{bucket}/o"),
                    params=dict(prefix=prefix, delimiter="/", maxResults="1"),
                    failure_exceptions={404: FileNotFoundError()},
                )
            )
            async with request.execute() as resp:
                result = await resp.json()
                return "items" in result or "prefixes" in result
        else:
            request = await googlify_request(
                Request(
                    method="GET",
                    url=path.format_url("https://storage.googleapis.com/storage/v1/b/{bucket}"),
                    failure_exceptions={404: FileNotFoundError()},
                )
            )
            async with request.execute() as resp:
                return True
    except FileNotFoundError:
        return False


@isdir.register  # type: ignore
async def _local_isdir(path: LocalPath) -> bool:
    return os.path.isdir(path)


# ==============================
# isfile
# ==============================


@pathdispatch
async def isfile(path: Union[BasePath, str]) -> bool:
    raise ValueError(f"Unsupported path: {path}")


@isfile.register  # type: ignore
async def _cloud_isfile(path: CloudPath) -> bool:
    try:
        await stat(path)
        return True
    except FileNotFoundError:
        return False


@isfile.register  # type: ignore
async def _local_isfile(path: LocalPath) -> bool:
    return os.path.isfile(path)


# ==============================
# exists
# ==============================


@pathdispatch
async def exists(path: Union[BasePath, str]) -> int:
    raise ValueError(f"Unsupported path: {path}")


@exists.register  # type: ignore
async def _cloud_exists(path: CloudPath) -> int:
    # TODO: this is two network requests, make it one
    for fut in asyncio.as_completed([isfile(path), isdir(path)]):
        if await fut:
            return True
    return False


@exists.register  # type: ignore
async def _local_exists(path: LocalPath) -> int:
    return os.path.exists(path)


# ==============================
# helpers
# ==============================


def url_format(template: str, **data: Any) -> str:
    escaped_data = {k: urllib.parse.quote(v, safe="") for k, v in data.items()}
    return template.format(**escaped_data)


def _strip_slash(path: str) -> str:
    return path[:-1] if path.endswith("/") else path


def _azure_parse_timestamp(text: str) -> float:
    return datetime.datetime.strptime(
        text.replace("GMT", "Z"), "%a, %d %b %Y %H:%M:%S %z"
    ).timestamp()


def _azure_get_md5(metadata: Mapping[str, Any]) -> Optional[str]:
    if "Content-MD5" in metadata:
        b64_encoded = metadata["Content-MD5"]
        if b64_encoded is None:
            return None
        return base64.b64decode(b64_encoded).hex()
    return None


def _google_parse_timestamp(text: str) -> float:
    return datetime.datetime.strptime(text, "%Y-%m-%dT%H:%M:%S.%f%z").timestamp()


def _google_get_md5(metadata: Mapping[str, Any]) -> Optional[str]:
    if "md5Hash" in metadata:
        return base64.b64decode(metadata["md5Hash"]).hex()

    if "metadata" in metadata and "md5" in metadata["metadata"]:
        # fallback to our custom hash if this is a composite object that is lacking the
        # md5Hash field
        return metadata["metadata"]["md5"]

    return None



import os
import sys

from boostedblob.cli import run_bbb


def main() -> None:
    try:
        run_bbb(sys.argv[1:])
    except Exception:
        if os.environ.get("BBB_DEBUG") or os.environ.get("BBB_TRACEBACK"):
            raise
        sys.exit(1)


if __name__ == "__main__":
    main()



from __future__ import annotations

import base64
import binascii
import datetime
import hashlib
import json
import os
import socket
import sys
import time
import urllib.parse
from typing import TYPE_CHECKING, Any, List, Mapping, Tuple

from Cryptodome.Hash import SHA256
from Cryptodome.PublicKey import RSA
from Cryptodome.Signature import pkcs1_15

if TYPE_CHECKING:
    from .path import GooglePath
    from .request import Request

MAX_EXPIRATION_SECONDS = 7 * 24 * 60 * 60


def default_gcloud_path() -> str:
    if sys.platform == "win32":
        # https://www.jhanley.com/google-cloud-application-default-credentials/
        return os.path.join(os.environ["APPDATA"], "gcloud")
    return os.path.join(os.environ["HOME"], ".config/gcloud")


def load_credentials() -> Mapping[str, Any]:
    if "GOOGLE_APPLICATION_CREDENTIALS" in os.environ:
        creds_path = os.environ["GOOGLE_APPLICATION_CREDENTIALS"]
        if not os.path.exists(creds_path):
            raise RuntimeError(
                f"Credentials not found at {creds_path} specified by environment variable "
                "'GOOGLE_APPLICATION_CREDENTIALS'"
            )
        with open(creds_path) as f:
            return json.load(f)

    default_creds_path = os.path.join(default_gcloud_path(), "application_default_credentials.json")
    if os.path.exists(default_creds_path):
        with open(default_creds_path) as f:
            return json.load(f)
    raise RuntimeError(
        "Credentials not found, please login with 'gcloud auth application-default login' or "
        "else set the 'GOOGLE_APPLICATION_CREDENTIALS' environment variable to the path of a "
        "JSON format service account key"
    )


async def get_access_token(_: str) -> Tuple[Any, float]:
    from .request import Request

    now = time.time()

    # https://github.com/googleapis/google-auth-library-java/blob/master/README.md#application-default-credentials
    try:
        creds = load_credentials()
    except RuntimeError:
        if os.environ.get("NO_GCE_CHECK", "false").lower() != "true" and _is_gce_instance():
            # see if the metadata server has a token for us
            req = Request(
                method="GET",
                url="http://metadata.google.internal/computeMetadata/v1/instance/service-accounts/default/token",
                headers={"Metadata-Flavor": "Google"},
            )
            async with req.execute() as resp:
                result = await resp.json()
            return result["access_token"], now + float(result["expires_in"])
        raise

    req = create_access_token_request(
        creds, scopes=["https://www.googleapis.com/auth/devstorage.full_control"]
    )
    req.success_codes = (200, 400)

    async with req.execute() as resp:
        result = await resp.json()
        if resp.status == 400:
            error = result["error"]
            description = result.get("error_description", "<missing description>")
            msg = f"Error with google credentials: [{error}] {description}"
            if error == "invalid_grant":
                if description.startswith("Invalid JWT:"):
                    msg += "\nPlease verify that your system clock is correct."
                elif description == "Bad Request":
                    msg += (
                        "\nYour credentials may be expired, please run the following commands: "
                        "`gcloud auth application-default revoke` (this may fail but ignore the "
                        "error) then `gcloud auth application-default login`"
                    )
            raise RuntimeError(msg)
        assert resp.status == 200
    return result["access_token"], now + float(result["expires_in"])


def create_access_token_request(creds: Mapping[str, Any], scopes: List[str]) -> Request:
    if "private_key" in creds:
        # looks like GCS does not support the no-oauth flow
        # https://developers.google.com/identity/protocols/OAuth2ServiceAccount#jwt-auth
        return create_token_request(creds["client_email"], creds["private_key"], scopes)
    if "refresh_token" in creds:
        return _refresh_access_token_request(
            refresh_token=creds["refresh_token"],
            client_id=creds["client_id"],
            client_secret=creds["client_secret"],
        )
    raise RuntimeError("Credentials not recognized")


def create_token_request(client_email: str, private_key: str, scopes: List[str]) -> Request:
    from .request import Request

    # https://developers.google.com/identity/protocols/OAuth2ServiceAccount
    now = time.time()
    claim_set = {
        "iss": client_email,
        "scope": " ".join(scopes),
        "aud": "https://www.googleapis.com/oauth2/v4/token",
        "exp": now + 60 * 60,
        "iat": now,
    }
    data = {
        "grant_type": "urn:ietf:params:oauth:grant-type:jwt-bearer",
        "assertion": _create_jwt(private_key, claim_set),
    }
    return Request(
        url="https://www.googleapis.com/oauth2/v4/token",
        method="POST",
        headers={"Content-Type": "application/x-www-form-urlencoded"},
        data=urllib.parse.urlencode(data).encode("utf8"),
    )


def _refresh_access_token_request(
    client_id: str, client_secret: str, refresh_token: str
) -> Request:
    from .request import Request

    # https://developers.google.com/identity/protocols/OAuth2WebServer#offline
    data = {
        "grant_type": "refresh_token",
        "refresh_token": refresh_token,
        "client_id": client_id,
        "client_secret": client_secret,
    }
    return Request(
        url="https://www.googleapis.com/oauth2/v4/token",
        method="POST",
        headers={"Content-Type": "application/x-www-form-urlencoded"},
        data=urllib.parse.urlencode(data).encode("utf8"),
    )


def _is_gce_instance() -> bool:
    try:
        socket.getaddrinfo("metadata.google.internal", 80)
    except socket.gaierror:
        return False
    return True


def _sign(private_key: str, msg: bytes) -> bytes:
    key = RSA.import_key(private_key)
    h = SHA256.new(msg)
    return pkcs1_15.new(key).sign(h)


def _create_jwt(private_key: str, data: Mapping[str, Any]) -> bytes:
    encode = base64.urlsafe_b64encode
    header_b64 = encode(json.dumps({"alg": "RS256", "typ": "JWT"}).encode("utf8"))
    body_b64 = encode(json.dumps(data).encode("utf8"))
    to_sign = header_b64 + b"." + body_b64
    signature_b64 = encode(_sign(private_key, to_sign))
    return header_b64 + b"." + body_b64 + b"." + signature_b64


def generate_signed_url(path: GooglePath) -> Tuple[str, datetime.datetime]:
    # https://cloud.google.com/storage/docs/access-control/signing-urls-manually
    creds = load_credentials()
    if "private_key" not in creds:
        raise RuntimeError(
            "Private key not found in credentials.  Please set the "
            "`GOOGLE_APPLICATION_CREDENTIALS` environment variable to point to a JSON key for "
            "a service account to use this call"
        )

    host = "storage.googleapis.com"
    canonical_uri = path.format_url("/{bucket}/{blob}")

    datetime_now = datetime.datetime.utcnow()
    request_timestamp = datetime_now.strftime("%Y%m%dT%H%M%SZ")
    datestamp = datetime_now.strftime("%Y%m%d")

    credential_scope = f"{datestamp}/auto/storage/goog4_request"
    credential = f"{creds['client_email']}/{credential_scope}"

    canonical_headers = ""
    ordered_headers = sorted({"host": host}.items())
    for k, v in ordered_headers:
        lower_k = str(k).lower()
        strip_v = str(v).lower()
        canonical_headers += f"{lower_k}:{strip_v}\n"

    signed_headers_parts = []
    for k, _ in ordered_headers:
        lower_k = str(k).lower()
        signed_headers_parts.append(lower_k)
    signed_headers = ";".join(signed_headers_parts)

    params = {
        "X-Goog-Algorithm": "GOOG4-RSA-SHA256",
        "X-Goog-Credential": credential,
        "X-Goog-Date": request_timestamp,
        "X-Goog-Expires": str(MAX_EXPIRATION_SECONDS),
        "X-Goog-SignedHeaders": signed_headers,
    }

    canonical_query_string_parts = []
    ordered_params = sorted(params.items())
    for k, v in ordered_params:
        encoded_k = urllib.parse.quote(str(k), safe="")
        encoded_v = urllib.parse.quote(str(v), safe="")
        canonical_query_string_parts.append(f"{encoded_k}={encoded_v}")
    canonical_query_string = "&".join(canonical_query_string_parts)

    canonical_request = "\n".join(
        [
            "GET",
            canonical_uri,
            canonical_query_string,
            canonical_headers,
            signed_headers,
            "UNSIGNED-PAYLOAD",
        ]
    )
    canonical_request_hash = hashlib.sha256(canonical_request.encode()).hexdigest()

    string_to_sign = "\n".join(
        ["GOOG4-RSA-SHA256", request_timestamp, credential_scope, canonical_request_hash]
    ).encode("utf8")
    signature = binascii.hexlify(_sign(creds["private_key"], string_to_sign)).decode("utf8")
    signed_url = f"https://{host}{canonical_uri}"
    signed_url += f"?{canonical_query_string}&X-Goog-Signature={signature}"
    return signed_url, datetime_now + datetime.timedelta(seconds=MAX_EXPIRATION_SECONDS)



from __future__ import annotations

import configparser
import datetime
import os
import re
from typing import Any, AsyncIterator, Iterator, Mapping, NamedTuple, Optional, Union

from . import google_auth
from .path import AzurePath, BasePath, CloudPath, GooglePath, LocalPath, Stat, isfile, pathdispatch
from .request import Request, azure_page_iterator, google_page_iterator


def format_size(num: float, suffix: str = "B") -> str:
    for unit in ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"]:
        if abs(num) < 1024.0:
            unit += suffix
            return f"{num:.1f} {unit:<3}"
        num /= 1024.0
    return f"{num:.1f} Yi{suffix}"


# ==============================
# DirEntry
# ==============================


class DirEntry(NamedTuple):
    path: BasePath
    is_dir: bool
    is_file: bool
    stat: Optional[Stat]

    @staticmethod
    def from_dirpath(path: BasePath) -> DirEntry:
        assert path.is_directory_like()
        return DirEntry(path=path, is_dir=True, is_file=False, stat=None)

    @staticmethod
    def from_path_stat(path: BasePath, stat: Stat) -> DirEntry:
        assert not path.is_directory_like()
        return DirEntry(path=path, is_dir=False, is_file=True, stat=stat)

    def format(self, human_readable: bool = False) -> str:
        size = (
            (format_size(self.stat.size) if human_readable else self.stat.size) if self.stat else ""
        )
        mtime = (
            datetime.datetime.fromtimestamp(int(self.stat.mtime)).isoformat() if self.stat else ""
        )
        return f"{size:>12}  {mtime:19}  {self.path}"


# ==============================
# list_blobs
# ==============================


@pathdispatch
def list_blobs(
    path: Union[CloudPath, str], delimiter: Optional[str], allow_prefix: bool = False
) -> AsyncIterator[DirEntry]:
    """List all blobs whose prefix matches ``path``.

    :param path: The prefix we want to list.
    :param delimiter: Group blobs matching the prefix up to a delimiter. This allows us to emulate
        directories on blob storage.
    :param allow_prefix: Whether to allow prefixes that do not end with a delimiter (or "/").

    """
    raise ValueError(f"Unsupported path: {path}")


@list_blobs.register  # type: ignore
async def _azure_list_blobs(
    path: AzurePath, delimiter: Optional[str], allow_prefix: bool = False
) -> AsyncIterator[DirEntry]:
    prefix = path
    if not allow_prefix:
        _delimiter = delimiter or "/"
        assert not prefix.blob or prefix.blob.endswith(_delimiter)
    if not path.container:
        if delimiter != "/":
            raise ValueError("Cannot list blobs in storage account; must specify container")
        async for entry in _azure_list_containers(prefix.account):
            yield entry
        return

    params = {}
    if delimiter is not None:
        params["delimiter"] = delimiter
    it = azure_page_iterator(
        Request(
            method="GET",
            url=prefix.format_url("https://{account}.blob.core.windows.net/{container}"),
            params=dict(comp="list", restype="container", prefix=prefix.blob, **params),
        )
    )

    async for result in it:
        for entry in _azure_get_entries(prefix.account, prefix.container, result):
            yield entry


@list_blobs.register  # type: ignore
async def _google_list_blobs(
    path: GooglePath, delimiter: Optional[str], allow_prefix: bool = False
) -> AsyncIterator[DirEntry]:
    prefix = path
    if not allow_prefix:
        _delimiter = delimiter or "/"
        assert not prefix.blob or prefix.blob.endswith(_delimiter)
    if not path.bucket:
        if delimiter != "/":
            raise ValueError("Cannot list blobs across buckets")
        async for entry in _google_list_buckets():
            yield entry
        return

    params = {}
    if delimiter is not None:
        params["delimiter"] = delimiter
    it = google_page_iterator(
        Request(
            method="GET",
            url=prefix.format_url("https://storage.googleapis.com/storage/v1/b/{bucket}/o"),
            params=dict(prefix=prefix.blob, **params),
        )
    )

    async for result in it:
        for entry in _google_get_entries(prefix.bucket, result):
            yield entry


# ==============================
# scandir
# ==============================


@pathdispatch
def scandir(path: Union[BasePath, str]) -> AsyncIterator[DirEntry]:
    """Iterate over the entries in a directory.

    :param path: The directory we want to scan.

    """
    raise ValueError(f"Unsupported path: {path}")


@scandir.register  # type: ignore
async def _cloud_scandir(path: CloudPath) -> AsyncIterator[DirEntry]:
    dirpath = path.ensure_directory_like()
    subpath_exists = False
    async for entry in list_blobs(dirpath, delimiter="/"):
        subpath_exists = True
        if entry.path == dirpath:
            continue
        yield entry

    # If we find nothing, then run some checks so we throw the appropriate error.
    # Doing this means we avoid extra requests in the happy path.
    if not subpath_exists:
        if not await isfile(path):
            raise FileNotFoundError(path)
        raise NotADirectoryError(path)


@scandir.register  # type: ignore
async def _local_scandir(path: LocalPath) -> AsyncIterator[DirEntry]:
    path_absolute = os.path.abspath(path)
    for entry in os.scandir(path_absolute):
        # entry.path is absolute since we're passing an absolute path to os.scandir
        entry_path = LocalPath(entry.path)
        if entry.is_dir():
            yield DirEntry(
                path=entry_path.ensure_directory_like(), is_dir=True, is_file=False, stat=None
            )
        else:
            yield DirEntry(path=entry_path, is_dir=False, is_file=True, stat=_os_safe_stat(entry))


# ==============================
# listdir
# ==============================


@pathdispatch
def listdir(path: Union[BasePath, str]) -> AsyncIterator[BasePath]:
    """Iterate over the names of entries in a directory.

    :param path: The directory we want to scan.

    """
    raise ValueError(f"Unsupported path: {path}")


@listdir.register  # type: ignore
async def _cloud_listdir(path: CloudPath) -> AsyncIterator[CloudPath]:
    async for entry in scandir(path):
        assert isinstance(entry.path, type(path))
        yield entry.path


@listdir.register  # type: ignore
async def _local_listdir(path: LocalPath) -> AsyncIterator[LocalPath]:
    for p in os.listdir(path):
        yield LocalPath(p)


# ==============================
# scantree
# ==============================


@pathdispatch
def scantree(path: Union[BasePath, str]) -> AsyncIterator[DirEntry]:
    """Iterate over file entries in the directory tree rooted at path.

    :param path: The root of the tree we want to scan.

    """
    raise ValueError(f"Unsupported path: {path}")


@scantree.register  # type: ignore
async def _cloud_scantree(path: CloudPath) -> AsyncIterator[DirEntry]:
    dirpath = path.ensure_directory_like()
    subpath_exists = False
    async for entry in list_blobs(dirpath, None):
        subpath_exists = True
        if entry.path == dirpath:
            continue
        yield entry

    # If we find nothing, then run some checks so we throw the appropriate error.
    # Doing this means we avoid extra requests in the happy path.
    if not subpath_exists:
        if not await isfile(path):
            raise FileNotFoundError(path)
        raise NotADirectoryError(path)


@scantree.register  # type: ignore
async def _local_scantree(path: LocalPath) -> AsyncIterator[DirEntry]:
    path_absolute = os.path.abspath(path)

    def inner(current: str) -> Iterator[DirEntry]:
        entries = list(os.scandir(current))
        for entry in entries:
            if entry.is_dir():
                yield from inner(entry.path)
            else:
                yield DirEntry(
                    path=LocalPath(entry.path),
                    is_dir=False,
                    is_file=True,
                    stat=_os_safe_stat(entry),
                )

    for entry in inner(path_absolute):
        yield entry


# ==============================
# listtree
# ==============================


@pathdispatch
def listtree(path: Union[BasePath, str]) -> AsyncIterator[BasePath]:
    """List files in the directory tree rooted at path.

    :param path: The directory we want to scan.

    """
    raise ValueError(f"Unsupported path: {path}")


@listtree.register  # type: ignore
async def _cloud_listtree(path: CloudPath) -> AsyncIterator[CloudPath]:
    async for entry in scantree(path):
        assert isinstance(entry.path, type(path))
        yield entry.path


@listtree.register  # type: ignore
async def _local_listtree(path: LocalPath) -> AsyncIterator[LocalPath]:
    for base, _subdirs, files in os.walk(path):
        for name in files:
            yield LocalPath(base) / name


# ==============================
# glob_scandir
# ==============================


@pathdispatch
def glob_scandir(path: Union[BasePath, str]) -> AsyncIterator[DirEntry]:
    """Iterate over a glob in a directory.

    :param path: The glob to match against.

    """
    raise ValueError(f"Unsupported path: {path}")


@glob_scandir.register  # type: ignore
async def _azure_glob_scandir(path: AzurePath) -> AsyncIterator[DirEntry]:
    if "*" in path.account:
        raise ValueError("Cannot use wildcard in storage account")
    if "*" in path.container:
        if path.blob:
            raise ValueError("Currently only supports wildcards inside the filename")
        pattern = _glob_to_regex(path.container)
        async for entry in _azure_list_containers(path.account):
            assert isinstance(entry.path, type(path))
            if re.match(pattern, entry.path.container):
                yield entry
        return
    if "*" in os.path.dirname(path.blob):
        raise ValueError("Currently only supports wildcards inside the filename")
    if "*" not in path.name:
        path = path / "*"
    pattern = _glob_to_regex(path.name)

    blob_prefix = path.blob.split("*", maxsplit=1)[0]
    prefix = AzurePath(account=path.account, container=path.container, blob=blob_prefix)
    async for entry in list_blobs(prefix, delimiter="/", allow_prefix=True):
        assert isinstance(entry.path, type(path))
        if entry.path == path.parent.ensure_directory_like():
            # skip directory file marker of the parent, if present
            continue
        if re.match(pattern, entry.path.name):
            yield entry


@glob_scandir.register  # type: ignore
async def _google_glob_scandir(path: GooglePath) -> AsyncIterator[DirEntry]:
    if "*" in path.bucket:
        if path.blob:
            raise ValueError("Currently only supports wildcards inside the filename")
        pattern = _glob_to_regex(path.bucket)
        async for entry in _google_list_buckets():
            assert isinstance(entry.path, type(path))
            if re.match(pattern, entry.path.bucket):
                yield entry
        return
    if "*" in os.path.dirname(path.blob):
        raise ValueError("Currently only supports wildcards inside the filename")
    if "*" not in path.name:
        path = path / "*"
    pattern = _glob_to_regex(path.name)

    blob_prefix = path.blob.split("*", maxsplit=1)[0]
    prefix = GooglePath(bucket=path.bucket, blob=blob_prefix)
    async for entry in list_blobs(prefix, delimiter="/", allow_prefix=True):
        assert isinstance(entry.path, type(path))
        if entry.path == path.parent.ensure_directory_like():
            # skip directory file marker of the parent, if present
            continue
        if re.match(pattern, entry.path.name):
            yield entry


@glob_scandir.register  # type: ignore
async def _local_glob_scandir(path: LocalPath) -> AsyncIterator[DirEntry]:
    if "*" in os.path.dirname(path):
        raise ValueError("Currently only supports wildcards inside the filename")
    if "*" not in path.name:
        path = path / "*"
    pattern = _glob_to_regex(path.name)

    async for entry in scandir(path.parent):
        if re.match(pattern, entry.path.name):
            yield entry


# ==============================
# helpers
# ==============================


def _azure_get_entries(
    account: str, container: str, result: Mapping[str, Any]
) -> Iterator[DirEntry]:
    blobs = result["Blobs"]
    if blobs is None:
        return
    if "BlobPrefix" in blobs:
        if isinstance(blobs["BlobPrefix"], dict):
            blobs["BlobPrefix"] = [blobs["BlobPrefix"]]
        for bp in blobs["BlobPrefix"]:
            blob = bp["Name"]
            path = AzurePath(account, container, blob)
            yield DirEntry.from_dirpath(path)
    if "Blob" in blobs:
        if isinstance(blobs["Blob"], dict):
            blobs["Blob"] = [blobs["Blob"]]
        for b in blobs["Blob"]:
            blob = b["Name"]
            path = AzurePath(account, container, blob)
            if b["Name"].endswith("/"):
                yield DirEntry.from_dirpath(path)
            else:
                props = b["Properties"]
                yield DirEntry.from_path_stat(path, Stat.from_azure(props))


def _google_get_entries(bucket: str, result: Mapping[str, Any]) -> Iterator[DirEntry]:
    if "prefixes" in result:
        for blob in result["prefixes"]:
            path = GooglePath(bucket, blob)
            yield DirEntry.from_dirpath(path)
    if "items" in result:
        for item in result["items"]:
            blob = item["name"]
            path = GooglePath(bucket, blob)
            if item["name"].endswith("/"):
                yield DirEntry.from_dirpath(path)
            else:
                yield DirEntry.from_path_stat(path, Stat.from_google(item))


async def _azure_list_containers(account: str) -> AsyncIterator[DirEntry]:
    it = azure_page_iterator(
        Request(
            method="GET",
            url=f"https://{account}.blob.core.windows.net/",
            params=dict(comp="list"),
            failure_exceptions={404: FileNotFoundError(AzurePath(account, "", ""))},
        )
    )
    async for result in it:
        result_containers = result["Containers"]
        if result_containers is None:
            raise ValueError(f"No containers found in storage account {account}")
        result_containers = result_containers["Container"]
        containers = (
            [result_containers] if isinstance(result_containers, dict) else result_containers
        )
        assert isinstance(containers, list)
        for container in containers:
            yield DirEntry(
                path=AzurePath(account, container["Name"], ""),
                is_dir=True,
                is_file=False,
                stat=None,
            )


async def _google_list_buckets(project: Optional[str] = None) -> AsyncIterator[DirEntry]:
    if project is None:
        try:
            config = configparser.ConfigParser()
            config.read(
                os.path.join(google_auth.default_gcloud_path(), "configurations/config_default")
            )
            project = config["core"]["project"]
        except (KeyError, FileNotFoundError):
            raise ValueError(
                "Could not determine project in which to list buckets; try setting it "
                "with `gsutil config`"
            )

    it = google_page_iterator(
        Request(
            method="GET",
            url="https://storage.googleapis.com/storage/v1/b",
            params=dict(project=project),
        )
    )
    async for result in it:
        buckets = result["items"]
        assert isinstance(buckets, list)
        for bucket in buckets:
            yield DirEntry(
                path=GooglePath(bucket["name"], ""), is_dir=True, is_file=False, stat=None
            )


def _glob_to_regex(pattern: str) -> str:
    tokens = (token for token in re.split("([*]+)", pattern) if token != "")
    regexp = ""
    for token in tokens:
        if token == "*":
            regexp += r"[^/]*"
        elif token == "**":
            regexp += r".*"
        else:
            regexp += re.escape(token)
    return regexp + r"/?$"


def _os_safe_stat(entry: os.DirEntry[Any]) -> Optional[Stat]:
    # stat can fail and we don't want that to affect listing operations
    # In particular, we ran into the equivalent of the following raising FileNotFoundError:
    # next(x for x in os.scandir("/") if x.name == ".VolumeIcon.icns").stat()
    try:
        return Stat.from_local(entry.stat())
    except Exception:
        return None



import argparse
import asyncio
import functools
import os
import sys
from typing import Any, AsyncIterator, Awaitable, Callable, Dict, List, TypeVar

import boostedblob as bbb

F = TypeVar("F", bound=Callable[..., Any])
T = TypeVar("T")


def syncify(fn: Callable[..., Awaitable[T]]) -> Callable[..., T]:
    @functools.wraps(fn)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        try:
            import uvloop

            uvloop.install()
        except ImportError:
            pass
        return asyncio.run(fn(*args, **kwargs))

    return wrapper


def sync_with_session(fn: F) -> F:
    return syncify(bbb.ensure_session(fn))  # type: ignore


DEFAULT_CONCURRENCY = 100


def is_glob(path: str) -> bool:
    return "*" in path


async def print_long(it: AsyncIterator[bbb.listing.DirEntry], human_readable: bool) -> None:
    total = 0
    num_files = 0
    async for entry in it:
        if entry.is_file:
            num_files += 1
        if entry.stat:
            total += entry.stat.size
        print(entry.format(human_readable=human_readable))
    if human_readable:
        human_total = bbb.listing.format_size(total).strip()
        print(f"Listed {num_files} files summing to {total} bytes ({human_total})")


@sync_with_session
async def ls(path: str, long: bool = False, machine: bool = False) -> None:
    path_obj = bbb.BasePath.from_str(path)
    if is_glob(path):
        it = bbb.listing.glob_scandir(path_obj)
        if long:
            await print_long(it, human_readable=not machine)
        else:
            async for entry in it:
                print(entry.path)
        return

    try:
        if long:
            await print_long(bbb.scandir(path_obj), human_readable=not machine)
        else:
            async for p in bbb.listdir(path_obj):
                print(p)
    except NotADirectoryError:
        if long:
            stat = await bbb.stat(path_obj)
            entry = bbb.listing.DirEntry.from_path_stat(path_obj, stat)
            print(entry.format(human_readable=not machine))
        else:
            print(path_obj)


@sync_with_session
async def lstree(path: str, long: bool = False, machine: bool = False) -> None:
    try:
        if long:
            await print_long(bbb.scantree(path), human_readable=not machine)
        else:
            async for p in bbb.listtree(path):
                print(p)
    except NotADirectoryError:
        path_obj = bbb.BasePath.from_str(path)
        if long:
            stat = await bbb.stat(path_obj)
            entry = bbb.listing.DirEntry.from_path_stat(path_obj, stat)
            print(entry.format(human_readable=not machine))
        else:
            print(path_obj)


@sync_with_session
async def cat(path: str, concurrency: int = DEFAULT_CONCURRENCY) -> None:
    loop = asyncio.get_event_loop()
    async with bbb.BoostExecutor(concurrency) as executor:
        stream = await bbb.read.read_stream(path, executor)
        async for data in bbb.boost.iter_underlying(stream):
            await loop.run_in_executor(None, sys.stdout.buffer.write, data)


@sync_with_session
async def cp(
    srcs: List[str], dst: str, quiet: bool = False, concurrency: int = DEFAULT_CONCURRENCY
) -> None:
    dst_obj = bbb.BasePath.from_str(dst)
    dst_is_dirlike = dst_obj.is_directory_like() or await bbb.isdir(dst_obj)

    async with bbb.BoostExecutor(concurrency) as executor:
        if len(srcs) > 1 and not dst_is_dirlike:
            raise NotADirectoryError(dst_obj)

        async def copy_wrapper(src: str) -> None:
            src_obj = bbb.BasePath.from_str(src)
            if is_glob(src):
                if not dst_is_dirlike:
                    raise NotADirectoryError(dst_obj)
                async for path in bbb.copying.copyglob_iterator(src_obj, dst_obj, executor):
                    if not quiet:
                        print(path)
                return

            dst_file_obj = dst_obj / src_obj.name if dst_is_dirlike else dst_obj
            await bbb.copyfile(src_obj, dst_file_obj, executor, overwrite=True)
            if not quiet:
                print(src_obj)

        await bbb.boost.consume(executor.map_unordered(copy_wrapper, iter(srcs)))


@sync_with_session
async def cptree(
    src: str, dst: str, quiet: bool = False, concurrency: int = DEFAULT_CONCURRENCY
) -> None:
    src_obj = bbb.BasePath.from_str(src)
    async with bbb.BoostExecutor(concurrency) as executor:
        async for p in bbb.copying.copytree_iterator(src_obj, dst, executor):
            if not quiet:
                print(p)


@sync_with_session
async def rm(paths: List[str], quiet: bool = False, concurrency: int = DEFAULT_CONCURRENCY) -> None:
    async with bbb.BoostExecutor(concurrency) as executor:

        async def remove_wrapper(path: str) -> None:
            path_obj = bbb.BasePath.from_str(path)
            if is_glob(path):
                async for p in bbb.delete.glob_remove(path_obj, executor):
                    if not quiet:
                        print(p)
                return
            await bbb.remove(path_obj)
            if not quiet:
                print(path_obj)

        await bbb.boost.consume(executor.map_unordered(remove_wrapper, iter(paths)))


@sync_with_session
async def rmtree(path: str, quiet: bool = False, concurrency: int = DEFAULT_CONCURRENCY) -> None:
    path_obj = bbb.BasePath.from_str(path)
    async with bbb.BoostExecutor(concurrency) as executor:
        if is_glob(path):
            # this will fail if the glob matches a directory, which is a little contra the spirit of
            # rmtree. but maybe the best way to do that (and least likely to result in accidents) is
            # through recursive wildcards
            async for p in bbb.delete.glob_remove(path_obj, executor):
                if not quiet:
                    print(p)
        elif isinstance(path_obj, bbb.CloudPath):
            async for p in bbb.delete.rmtree_iterator(path_obj, executor):
                if not quiet:
                    print(p)
        else:
            await bbb.rmtree(path_obj, executor)


@sync_with_session
async def share(path: str) -> None:
    url, expiration = await bbb.share.get_url(path)
    print(url)
    if expiration is not None:
        print(f"Expires on: {expiration.isoformat()}")


@sync_with_session
async def sync(
    src: str,
    dst: str,
    delete: bool = False,
    quiet: bool = False,
    concurrency: int = DEFAULT_CONCURRENCY,
) -> None:
    src_obj = bbb.BasePath.from_str(src)
    dst_obj = bbb.BasePath.from_str(dst)

    src_is_dirlike = src_obj.is_directory_like() or await bbb.isdir(src_obj)
    if not src_is_dirlike:
        raise ValueError(f"{src_obj} is not a directory")
    async with bbb.BoostExecutor(concurrency) as executor:
        async for p in bbb.sync(src_obj, dst_obj, executor, delete=delete):
            if not quiet:
                print(p)


def complete_init(shell: str) -> None:
    if shell == "zsh":
        # zsh uses index-1 based arrays, so adjust CURRENT
        # the (f) things splits the output on newlines into an array
        # -U disables prefix matching
        # -S '' prevents zsh from inserting a space after the completion
        init_script = """\
_bbb_complete() {
    compadd -U  -S '' ${(f)"$(bbb complete command zsh $(( $CURRENT - 1)) $words)"}
}
compdef _bbb_complete bbb
"""
    elif shell == "bash":
        # use COMP_LINE instead of COMP_WORDS because bash uncustomisably splits words on colons
        # -o nospace prevents bash from inserting a space after the completion
        # life would be simpler if we did COMP_WORDBREAKS=${COMP_WORDBREAKS//:}
        init_script = """\
_bbb_complete() {
    local completions="$(bbb complete command bash $COMP_POINT "$COMP_LINE")"
    COMPREPLY=( $(compgen -W "$completions" '') )
}
complete -o nospace -F _bbb_complete bbb
"""
    else:
        raise ValueError(f"Unrecognised shell {shell}")
    print(init_script)


@sync_with_session
async def complete_command(shell: str, index: int, partial_command: List[str]) -> None:
    if shell == "bash":
        # the entire command is passed as a single argument and index is an index into that string
        assert len(partial_command) == 1
        command_str = partial_command[0]
        partial_command = command_str.split()
        if index == len(command_str):
            index = len(partial_command) - (0 if command_str[index - 1] == " " else 1)
        else:
            return  # TODO: support bash completion in the middle of a command

    if index <= 1:
        # TODO: support completion of subcommands
        return

    # assume we're trying to complete a path; just add a wildcard
    word_to_complete = partial_command[index] if index < len(partial_command) else ""
    path_to_complete = word_to_complete.lstrip("'\"")
    if not path_to_complete.endswith("*"):
        path_to_complete += "*"

    try:
        async for entry in bbb.listing.glob_scandir(path_to_complete):
            # bash won't let you complete before a colon without setting COMP_WORDBREAKS globally
            # instead, we just make sure our completion matches the part before the colon
            # to do that, we special case az:// paths
            if isinstance(entry.path, bbb.AzurePath) and path_to_complete.startswith("az://"):
                path_str = entry.path.to_az_str()
            else:
                path_str = str(entry.path)

            # bash won't split on a colon inside a quoted argument, though
            if shell == "bash" and not word_to_complete.startswith(("'", '"')):
                try:
                    path_str = path_str[path_str.index(":") + 1 :]
                except ValueError:
                    pass
            print(path_str)
    except Exception:
        # ignore errors, usually
        if os.environ.get("BBB_DEBUG"):
            raise


def parse_options(args: List[str]) -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument("--version", action="version", version=f"boostedblob {bbb.__version__}")
    subparsers = parser.add_subparsers(required=True)

    concurrency_kwargs: Dict[str, Any] = dict(
        type=int,
        metavar="N",
        default=DEFAULT_CONCURRENCY,
        help="Number of concurrent requests to use",
    )

    ls_desc = """\
`bbb ls` lists the immediate contents of a directory (both files and
subdirectories).

To see the contents of the current directory:
$ bbb ls .

To additionally see size and mtime of files in the current directory:
$ bbb ls -l .

To check whether a file exists:
$ bbb ls schrodingers_file.txt

Aliases:
bbb ll == bbb ls -l
"""
    lstree_desc = """\
`bbb lstree` lists all files present anywhere in the given directory tree.
Note that `bbb lstree` only lists files and will not list subdirectories
(unless they are marked by marker files).

To see all files in your bucket:
$ bbb lstree gs://my_bucket/

To additionally see size and mtime of files:
$ bbb lstree -l gs://my_bucket/

To make output more easily machine readable:
$ bbb lstree --machine .

Aliases:
bbb lsr == bbb lstree
bbb llr == bbb lstree -l == bbb lsr -l == bbb du
"""
    cat_desc = """\
`bbb cat` copies the contents of a given file to stdout.

Example:
$ bbb cat boostedblob/boost.py
"""
    cp_desc = """\
`bbb cp` copies a file to another location. If the location is a directory, the
file will be copied there under the same name, otherwise it will be copied to
the name provided. If the destination file already exists, it will be
overwritten.

Copy a file into a directory:
$ bbb cp frogs.txt my_directory

Copy multiple files into a directory:
$ bbb cp frogs.txt worms.txt my_directory/

Copy a file to a file with a different name:
$ bbb cp frogs.txt renamed_frogs.txt
"""
    cptree_desc = """\
`bbb cptree` copies an entire directory tree. The rule here is simple: provide
two directory paths, and bbb will create a copy of the files with the same
structure under the destination directory. If any of the files already exist,
they will be overwritten.

Create an exact copy of a directory somewhere else:
$ bbb cptree boostedblob gs://tmp/boostedblob

Aliases:
bbb cpr == bbb cptree
"""
    rm_desc = """\
`bbb rm` deletes the given files.

Example:
$ bbb rm frog.txt worm.txt
"""
    rmtree_desc = """\
`bbb rmtree` deletes an entire directory tree.

Example:
$ bbb rmtree boostedblob
"""
    share_desc = """\
`bbb share` prints a link you can use to open a file in a browser.

Example:
$ bbb share gs://bucket/frogs.txt
"""
    sync_desc = """\
`bbb sync` synchronises two directory trees. Provide two directory paths, and
bbb will change the destination so that it better mirrors the source.
Specifically, bbb will copy over or replace files in the destination that are
missing or have changed in the source. If --delete is specified, bbb will delete
files in the destination that are not present in the source.

Example:
$ bbb sync gs://tmp/boostedblob boostedblob

Using --delete and re-syncing will delete spurious_file.txt:
$ touch boostedblob/spurious_file.txt
$ bbb sync --delete gs://tmp/boostedblob boostedblob
"""
    complete_desc = """\
To enable tab completion for bash, add the following to your bashrc:
eval "$(bbb complete init bash)"

To enable tab completion for zsh, add the following to your zshrc:
eval "$(bbb complete init zsh)"
"""

    subparser = subparsers.add_parser(
        "ls",
        help="List contents of a directory",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=ls_desc,
    )
    subparser.set_defaults(command=ls)
    subparser.add_argument("path", help="Path of directory to list")
    subparser.add_argument(
        "-l", "--long", action="store_true", help="List information about each file"
    )
    subparser.add_argument(
        "--machine",
        action="store_true",
        help="Make output more easily machine readable",
    )

    subparser = subparsers.add_parser(
        "ll", formatter_class=argparse.RawDescriptionHelpFormatter, description=ls_desc
    )
    subparser.set_defaults(command=functools.partial(ls, long=True))
    subparser.add_argument("path", help="Path of directory to list")
    subparser.add_argument(
        "--machine",
        action="store_true",
        help="Make output more easily machine readable",
    )

    subparser = subparsers.add_parser(
        "lstree",
        aliases=["lsr"],
        help="List all files in a directory tree",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=lstree_desc,
    )
    subparser.set_defaults(command=lstree)
    subparser.add_argument("path", help="Root of directory tree to list")
    subparser.add_argument(
        "-l", "--long", action="store_true", help="List information about each file"
    )
    subparser.add_argument(
        "--machine",
        action="store_true",
        help="Make output more easily machine readable",
    )

    subparser = subparsers.add_parser(
        "llr",
        aliases=["du"],
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=lstree_desc,
    )
    subparser.set_defaults(command=functools.partial(lstree, long=True))
    subparser.add_argument("path", help="Root of directory tree to list")
    subparser.add_argument(
        "--machine",
        action="store_true",
        help="Make output more easily machine readable",
    )

    subparser = subparsers.add_parser(
        "cat",
        help="Print the contents of a file",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=cat_desc,
    )
    subparser.set_defaults(command=cat)
    subparser.add_argument("path", help="File whose contents to print")
    subparser.add_argument("--concurrency", **concurrency_kwargs)

    subparser = subparsers.add_parser(
        "cp",
        help="Copy files",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=cp_desc,
    )
    subparser.set_defaults(command=cp)
    subparser.add_argument("srcs", nargs="+", help="File(s) to copy from")
    subparser.add_argument("dst", help="File or directory to copy to")
    subparser.add_argument("-q", "--quiet", action="store_true")
    subparser.add_argument("--concurrency", **concurrency_kwargs)

    subparser = subparsers.add_parser(
        "cptree",
        aliases=["cpr"],
        help="Copy a directory tree",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=cptree_desc,
    )
    subparser.set_defaults(command=cptree)
    subparser.add_argument("src", help="Directory to copy from")
    subparser.add_argument("dst", help="Directory to copy to")
    subparser.add_argument("-q", "--quiet", action="store_true")
    subparser.add_argument("--concurrency", **concurrency_kwargs)

    subparser = subparsers.add_parser(
        "rm",
        help="Remove files",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=rm_desc,
    )
    subparser.set_defaults(command=rm)
    subparser.add_argument("paths", nargs="+", help="File(s) to delete")
    subparser.add_argument("-q", "--quiet", action="store_true")
    subparser.add_argument("--concurrency", **concurrency_kwargs)

    subparser = subparsers.add_parser(
        "rmtree",
        aliases=["rmr"],
        help="Remove a directory tree",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=rmtree_desc,
    )
    subparser.set_defaults(command=rmtree)
    subparser.add_argument("path", help="Directory to delete")
    subparser.add_argument("-q", "--quiet", action="store_true")
    subparser.add_argument("--concurrency", **concurrency_kwargs)

    subparser = subparsers.add_parser(
        "share",
        help="Get a shareable link to a file",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=share_desc,
    )
    subparser.set_defaults(command=share)
    subparser.add_argument("path", help="Path to share")

    subparser = subparsers.add_parser(
        "sync",
        help="Sync a directory tree",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=sync_desc,
    )
    subparser.set_defaults(command=sync)
    subparser.add_argument("src", help="Directory to sync from")
    subparser.add_argument("dst", help="Directory to sync to")
    subparser.add_argument(
        "--delete", action="store_true", help="Delete destination files that don't exist in source"
    )
    subparser.add_argument("-q", "--quiet", action="store_true")
    subparser.add_argument("--concurrency", **concurrency_kwargs)

    subparser = subparsers.add_parser(
        "complete",
        help="Tab complete a command",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=complete_desc,
    )
    subsubparsers = subparser.add_subparsers(required=True)
    subsubparser = subsubparsers.add_parser(
        "init", help="Print shell script to initialise tab completion"
    )
    subsubparser.set_defaults(command=complete_init)
    subsubparser.add_argument("shell", choices=["bash", "zsh"])

    subsubparser = subsubparsers.add_parser("command", help="Get a completion for a command")
    subsubparser.set_defaults(command=complete_command)
    subsubparser.add_argument("shell", help="Shell to complete for")
    subsubparser.add_argument("index", type=int, help="Index into partial_command to complete")
    subsubparser.add_argument(
        "partial_command", nargs=argparse.REMAINDER, help="Command to complete"
    )

    if not args:
        parser.print_help()
        print()
        parser.error("missing subcommand, see `bbb --help`")
    return parser.parse_args(args)


def run_bbb(argv: List[str]) -> None:
    try:
        args = parse_options(argv)
        command = args.__dict__.pop("command")
        command(**args.__dict__)
    except Exception as e:
        print(f"ERROR: {type(e).__name__}: {e}", file=sys.stderr)
        raise



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['boostedblob']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.7.2,<4.0.0',
 'pycryptodomex>=3.9,<4.0',
 'uvloop>=0.14,<0.15',
 'xmltodict>=0.12,<0.13']

entry_points = \
{'console_scripts': ['bbb = boostedblob.__main__:main']}

setup_kwargs = {
    'name': 'boostedblob',
    'version': '0.5.0',
    'description': 'Command line tool and async library to perform basic file operations on local paths, Google Cloud Storage paths and Azure Blob Storage paths.',
    'long_description': '# boostedblob\n\nboostedblob is a command line tool and async library to perform basic file operations on local\npaths, Google Cloud Storage paths and Azure Blob Storage paths.\n\nboostedblob is derived from the excellent [blobfile](https://github.com/christopher-hesse/blobfile).\n\nThe fun part of implementing boostedblob is `boostedblob/boost.py`, which provides a\n`concurrent.futures`-like interface for running and composing async tasks in a concurrency limited\nenvironment.\n\n## Installation\n\nJust run `pip install boostedblob`. boostedblob requires Python 3.7 or better.\n\n<sup>(For Python 3.9 support, you\'ll need to run `pip uninstall uvloop`)</sup>\n\n\n## Usage\n\nFor an overview and list of commands:\n```sh\nbbb --help\n```\n\nFor help with a specific command:\n```sh\nbbb ls --help\n```\n\nTo enable tab completion, add the following to your shell config (replacing `zsh` with `bash`, if appropriate):\n```sh\neval "$(bbb complete init zsh)"\n```\nNote that the quotes are necessary. You can also inline the result of `bbb complete init zsh` into\nyour shell config to make your shell startup a little faster.\n\n## Contributing\n\nTo get started with developing boostedblob and running tests, see [CONTRIBUTING.md](CONTRIBUTING.md)\n',
    'author': 'Shantanu Jain',
    'author_email': 'hauntsaninja@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import datetime
import os
from typing import Optional, Tuple, Union

from . import azure_auth, google_auth
from .path import AzurePath, BasePath, GooglePath, LocalPath, pathdispatch


@pathdispatch
async def get_url(path: Union[BasePath, str]) -> Tuple[str, Optional[datetime.datetime]]:
    raise ValueError(f"Unsupported path: {path}")


@get_url.register  # type: ignore
async def _azure_get_url(path: AzurePath) -> Tuple[str, datetime.datetime]:
    return await azure_auth.generate_signed_url(path)


@get_url.register  # type: ignore
async def _google_get_url(path: GooglePath) -> Tuple[str, datetime.datetime]:
    return google_auth.generate_signed_url(path)


@get_url.register  # type: ignore
async def _local_get_url(path: LocalPath) -> Tuple[str, None]:
    return f"file://{os.path.abspath(path)}", None



__version__ = "0.5.0"

from . import read as read
from . import share as share
from . import write as write
from .boost import BoostExecutor as BoostExecutor
from .copying import copyfile as copyfile
from .copying import copytree as copytree
from .delete import remove as remove
from .delete import rmtree as rmtree
from .globals import ensure_session as ensure_session
from .listing import listdir as listdir
from .listing import listtree as listtree
from .listing import scandir as scandir
from .listing import scantree as scantree
from .path import AzurePath as AzurePath
from .path import BasePath as BasePath
from .path import CloudPath as CloudPath
from .path import GooglePath as GooglePath
from .path import LocalPath as LocalPath
from .path import exists as exists
from .path import getsize as getsize
from .path import isdir as isdir
from .path import isfile as isfile
from .path import stat as stat
from .syncing import sync as sync



from __future__ import annotations

import asyncio
import contextlib
import functools
import os
import sys
import time
from dataclasses import dataclass
from typing import (
    Any,
    AsyncIterator,
    Awaitable,
    Callable,
    Dict,
    Generic,
    Iterator,
    Optional,
    Tuple,
    TypeVar,
)

import aiohttp

from . import azure_auth, google_auth

MB = 2 ** 20


T = TypeVar("T")
F = TypeVar("F", bound=Callable[..., Any])


class TokenManager(Generic[T]):
    """Automatically refresh a token when it expires."""

    def __init__(self, get_token_fn: Callable[[T], Awaitable[Tuple[Any, float]]]) -> None:
        self._get_token_fn = get_token_fn
        self._tokens: Dict[T, Any] = {}
        self._expirations: Dict[T, float] = {}

    async def get_token(self, key: T) -> Any:
        now = time.time()
        expiration = self._expirations.get(key)
        if expiration is None or (now + config.token_early_expiration_seconds) > expiration:
            self._tokens[key], self._expirations[key] = await self._get_token_fn(key)
            assert self._expirations[key] is not None

        assert key in self._tokens
        return self._tokens[key]


@dataclass
class Config:
    session: Optional[aiohttp.ClientSession] = None

    debug_mode: bool = bool(os.environ.get("BBB_DEBUG"))

    chunk_size: int = 16 * MB

    connect_timeout: float = 30.0
    read_timeout: float = 300.0
    backoff_initial: float = 0.1
    backoff_max: float = 60.0
    backoff_jitter_fraction: float = 0.9
    retry_limit: int = 15

    token_early_expiration_seconds: int = 300
    azure_access_token_manager = TokenManager[Tuple[str, Optional[str]]](
        azure_auth.get_access_token
    )
    azure_sas_token_manager = TokenManager[Tuple[str, Optional[str]]](azure_auth.get_sas_token)
    google_access_token_manager = TokenManager[str](google_auth.get_access_token)


config: Config = Config()


@contextlib.contextmanager
def configure(**kwargs: Any) -> Iterator[None]:
    original = {k: getattr(config, k) for k in kwargs}
    config.__dict__.update(**kwargs)
    try:
        yield
    finally:
        config.__dict__.update(**original)


@contextlib.asynccontextmanager
async def session_context() -> AsyncIterator[None]:
    if config.session is None:
        set_event_loop_exception_handler()  # there could be a better place for this
        connector = aiohttp.TCPConnector(limit=0)
        async with aiohttp.ClientSession(connector=connector) as session:
            with configure(session=session):
                yield
    else:
        yield


def ensure_session(fn: F) -> F:
    @functools.wraps(fn)
    async def wrapper(*args: Any, **kwargs: Any) -> Any:
        async with session_context():
            return await fn(*args, **kwargs)

    return wrapper  # type: ignore


def set_event_loop_exception_handler() -> None:
    loop = asyncio.get_event_loop()

    def handler(
        loop: asyncio.AbstractEventLoop, context: Dict[str, Any]
    ) -> None:  # pragma: no cover
        message = context["message"]
        exception = context.get("exception", Exception)
        if sys.version_info < (3, 7, 4) and message in (
            "SSL error in data received",
            "Fatal error on transport",
        ):
            # Ignore aiohttp #3535 / cpython #13548 issue with SSL data after close
            # Adapted from https://github.com/aio-libs/aiohttp/issues/3535#issuecomment-483268542
            import ssl
            from asyncio.sslproto import SSLProtocol

            protocol = context.get("protocol")
            if (
                isinstance(exception, ssl.SSLError)
                and exception.reason == "KRB5_S_INIT"
                and isinstance(protocol, SSLProtocol)
            ):
                return

        # Our main interest here is minimising the other various errors and tracebacks that
        # drown out the politely formatted errors from cli.py when things go wrong
        if "exception was never retrieved" in message:
            from .request import MissingSession

            # While closing down, we remove the global session, causing other requests to fail. This
            # just causes noise, so ignore if that's the exception.
            if not isinstance(exception, MissingSession):
                print(
                    f"ERROR (while closing down): {type(exception).__name__}: {exception}",
                    file=sys.stderr,
                )
        else:
            print(
                f"ERROR (from event loop): {type(exception).__name__}: {message}", file=sys.stderr
            )
        if loop.get_debug():
            loop.default_exception_handler(context)

    loop.set_exception_handler(handler)
